"""
Module handling HTTP Requests and Connection Diagnostics
"""
from __future__ import print_function
from __future__ import absolute_import
import requests
import os
import six
import json
import logging
import platform
import xml.etree.ElementTree as ET
import warnings
import errno
# import io
from tempfile import TemporaryFile
# from datetime import datetime, timedelta
try:
    # python 2
    from urlparse import urlparse
    from urllib import quote
except ImportError:
    # python 3
    from urllib.parse import urlparse
    from urllib.parse import quote
from .utilities import (determine_hostname,
                        generate_machine_id,
                        machine_id_exists,
                        write_unregistered_file,
                        write_registered_file,
                        os_release_info,
                        largest_spec_in_archive,
                        size_in_mb)
from .cert_auth import rhsmCertificate
from .constants import InsightsConstants as constants
from insights import package_info
from insights.client.collection_rules import InsightsUploadConf
from insights.core import spec_cleaner
from insights.util.canonical_facts import get_canonical_facts

warnings.simplefilter('ignore')
APP_NAME = constants.app_name
NETWORK = constants.custom_network_log_level
logger = logging.getLogger(__name__)

"""
urllib3's logging is chatty
"""
URLLIB3_LOGGER = logging.getLogger('urllib3.connectionpool')
URLLIB3_LOGGER.setLevel(logging.WARNING)
URLLIB3_LOGGER = logging.getLogger('requests.packages.urllib3.connectionpool')
URLLIB3_LOGGER.setLevel(logging.WARNING)

REQUEST_FAILED_EXCEPTIONS = (requests.ConnectionError, requests.Timeout)

# TODO: Document this, or turn it into a real option
if os.environ.get('INSIGHTS_DEBUG_HTTP'):
    import httplib
    httplib.HTTPConnection.debuglevel = 1
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


def _host_not_found():
    raise Exception("Error: failed to find host with matching machine-id. Run insights-client --status to check registration status")


def _api_request_failed(exception, message='The Insights API could not be reached.'):
    logger.error(exception)
    if message:
        logger.error(message)


class InsightsConnection(object):

    """
    Helper class to manage details about the connection
    """

    def __init__(self, config):
        self.config = config
        self.username = self.config.username
        self.password = self.config.password

        # workaround while we support both legacy and plat APIs
        self.cert_verify = self.config.cert_verify
        if self.cert_verify is None:
            # if self.config.legacy_upload:
            self.cert_verify = os.path.join(
                constants.default_conf_dir,
                'cert-api.access.redhat.com.pem')
            # else:
            # self.cert_verify = True
        else:
            if isinstance(self.cert_verify, six.string_types):
                if self.cert_verify.lower() == 'false':
                    self.cert_verify = False
                elif self.cert_verify.lower() == 'true':
                    self.cert_verify = True

        protocol = "https://"

        self.auto_config = self.config.auto_config

        # workaround while we support both legacy and plat APIs
        # hack to "guess" the correct base URL if autoconfig off +
        #   no base_url in config
        if self.config.base_url is None:
            if self.config.legacy_upload:
                self.base_url = protocol + constants.legacy_base_url
            else:
                self.base_url = protocol + constants.base_url
        else:
            self.base_url = protocol + self.config.base_url
        # end hack. in the future, make console.redhat.com the default

        self.upload_url = self.config.upload_url
        if self.upload_url is None:
            if self.config.legacy_upload:
                self.upload_url = self.base_url + "/uploads"
            else:
                self.upload_url = self.base_url + '/ingress/v1/upload'

        self.api_url = self.base_url
        self.branch_info_url = self.config.branch_info_url
        if self.branch_info_url is None:
            # workaround for a workaround for a workaround
            base_url_base = self.base_url.split('/platform')[0]
            self.branch_info_url = base_url_base + '/v1/branch_info'
        self.inventory_url = self.api_url + "/inventory/v1"

        self.authmethod = self.config.authmethod
        self.systemid = self.config.systemid or None
        self.get_proxies()
        self.session = self._init_session()

    def _init_session(self):
        """
        Set up the session, auth is handled here
        """
        session = requests.Session()
        session.headers = {'User-Agent': self.user_agent,
                           'Accept': 'application/json'}
        if self.systemid is not None:
            session.headers.update({'systemid': self.systemid})
        if self.authmethod == "BASIC":
            session.auth = (self.username, self.password)
        elif self.authmethod == "CERT":
            cert = rhsmCertificate.certpath()
            key = rhsmCertificate.keypath()
            if rhsmCertificate.exists():
                session.cert = (cert, key)
            else:
                logger.error('ERROR: Certificates not found.')
        session.verify = self.cert_verify
        session.proxies = self.proxies
        session.trust_env = False
        return session

    def _http_request(self, url, method, log_response_text=True, **kwargs):
        '''
        Perform an HTTP request, net logging, and error handling
        Parameters
            url     - URL to perform the request against
            method  - HTTP method, used for logging
            kwargs  - Rest of the args to pass to the request function
        Returns
            HTTP response object
        '''
        logger.log(NETWORK, "%s %s", method, url)
        try:
            res = self.session.request(url=url, method=method, timeout=self.config.http_timeout, **kwargs)
        except Exception:
            raise
        logger.log(NETWORK, "HTTP Status: %d %s", res.status_code, res.reason)
        if log_response_text or res.status_code != 200:
            logger.log(NETWORK, "HTTP Response Text: %s", res.text)
        return res

    def get(self, url, **kwargs):
        try:
            return self._http_request(url, 'GET', **kwargs)
        except Exception:
            raise

    def post(self, url, **kwargs):
        try:
            return self._http_request(url, 'POST', **kwargs)
        except Exception:
            raise

    def put(self, url, **kwargs):
        try:
            return self._http_request(url, 'PUT', **kwargs)
        except Exception:
            raise

    def patch(self, url, **kwargs):
        try:
            return self._http_request(url, 'PATCH', **kwargs)
        except Exception:
            raise

    def delete(self, url, **kwargs):
        try:
            return self._http_request(url, 'DELETE', **kwargs)
        except Exception:
            raise

    @property
    def user_agent(self):
        """
        Generates and returns a string suitable for use as a request user-agent
        """
        import pkg_resources
        core_version = "insights-core"
        pkg = pkg_resources.working_set.find(pkg_resources.Requirement.parse(core_version))
        if pkg is not None:
            core_version = "%s %s" % (pkg.project_name, pkg.version)
        else:
            core_version = "Core %s" % package_info["VERSION"]

        try:
            from insights_client import constants as insights_client_constants
            client_version = "insights-client/{0}".format(insights_client_constants.InsightsConstants.version)
        except ImportError:
            client_version = "insights-client"

        if os.path.isfile(constants.ppidfile):
            with open(constants.ppidfile, 'r') as f:
                parent_process = f.read()
        else:
            parent_process = "unknown"

        requests_version = None
        pkg = pkg_resources.working_set.find(pkg_resources.Requirement.parse("requests"))
        if pkg is not None:
            requests_version = "%s %s" % (pkg.project_name, pkg.version)

        python_version = "%s %s" % (platform.python_implementation(), platform.python_version())

        os_family, os_release = os_release_info()
        kernel_version = "%s %s" % (platform.system(), platform.release())

        ua = "{client_version} ({core_version}; {requests_version}) {os_family} {os_release} ({python_version}; {kernel_version}); {parent_process}".format(
            client_version=client_version,
            core_version=core_version,
            parent_process=parent_process,
            python_version=python_version,
            os_family=os_family,
            os_release=os_release,
            kernel_version=kernel_version,
            requests_version=requests_version,
        )

        return ua

    def get_proxy(self, proxy_info, no_proxy_info, environment):
        proxies = None
        proxy_url = proxy_info
        if '@' in proxy_info:
            scheme = proxy_info.split(':')[0] + '://'
            logger.debug("Proxy Scheme: %s", scheme)
            location = proxy_info.split('@')[1]
            logger.debug("Proxy Location: %s", location)
            username = proxy_info.split(
                '@')[0].split(':')[1].replace('/', '')
            logger.debug("Proxy User: %s", username)
            proxy_url = proxy_info
            proxy_info = scheme + location
        logger.debug("%s Proxy: %s", environment, proxy_info)
        proxies = {"https": proxy_url}
        if no_proxy_info:
            insights_service_host = urlparse(self.base_url).hostname
            logger.debug('Found NO_PROXY set. Checking NO_PROXY %s against base URL %s.', no_proxy_info, insights_service_host)
            # Split the no_proxy entries on ',', then strip any leading and trailing whitespace.  Create a clean list for the
            # for loop.
            no_proxy_info = [host.strip() for host in no_proxy_info.split(',')]
            for no_proxy_host in no_proxy_info:
                logger.debug('Checking %s against %s', no_proxy_host, insights_service_host)
                if no_proxy_host == '*':
                    proxies = None
                    logger.debug('Found NO_PROXY asterisk(*) wildcard, disabling all proxies.')
                    break
                elif no_proxy_host.startswith('.') or no_proxy_host.startswith('*'):
                    if insights_service_host.endswith(no_proxy_host.replace('*', '')):
                        proxies = None
                        logger.debug('Found NO_PROXY range %s matching %s', no_proxy_host, insights_service_host)
                        break
                elif no_proxy_host == insights_service_host:
                    proxies = None
                    logger.debug('Found NO_PROXY %s exactly matching %s', no_proxy_host, insights_service_host)
                    break
        return proxies

    def get_proxies(self):
        """
        Determine proxy configuration
        """
        # Get proxy from ENV or Config
        proxies = None

        # CONF PROXY TAKES PRECEDENCE OVER ENV PROXY
        conf_proxy = self.config.proxy
        conf_no_proxy = self.config.no_proxy

        if conf_proxy:
            proxies = self.get_proxy(conf_proxy, conf_no_proxy, "CONF")

        # HANDLE NO PROXY CONF PROXY EXCEPTION VERBIAGE
        no_proxy = os.environ.get('NO_PROXY')
        if no_proxy and conf_proxy:
            logger.debug("You have environment variable NO_PROXY set "
                         "as well as 'proxy' set in your configuration file. "
                         "NO_PROXY environment variable will be ignored.")

        # IF NO CONF PROXY and NO_PROX none in conf, GET ENV PROXY AND NO PROXY
        if proxies is None and conf_no_proxy is None:
            env_proxy = os.environ.get('HTTPS_PROXY')
            if env_proxy:
                proxies = self.get_proxy(env_proxy, no_proxy, "ENV")
        self.proxies = proxies

    def _legacy_test_urls(self, url, method):
        """
        Actually test the url
        """
        # tell the api we're just testing the URL
        test_flag = {'test': 'test'}
        url = urlparse(url)
        test_url = url.scheme + "://" + url.netloc
        last_ex = None
        paths = (url.path + '/', '', '/r', '/r/insights')
        for ext in paths:
            try:
                logger.log(NETWORK, "Testing: %s", test_url + ext)
                if method == "POST":
                    test_req = self.post(test_url + ext, data=test_flag)
                elif method == "GET":
                    test_req = self.get(test_url + ext)
                # Strata returns 405 on a GET sometimes, this isn't a big deal
                if test_req.status_code in (200, 201):
                    logger.info(
                        "Successfully connected to: %s", test_url + ext)
                    return True
                else:
                    logger.info("Connection failed")
                    return False
            except requests.ConnectionError as exc:
                last_ex = exc
                logger.error(
                    "Could not successfully connect to: %s", test_url + ext)
                print(exc)
        if last_ex:
            raise last_ex

    def _test_urls(self, url, method):
        '''
        Test a URL
        '''
        if self.config.legacy_upload:
            return self._legacy_test_urls(url, method)
        try:
            logger.log(NETWORK, 'Testing %s', url)
            if method == 'POST':
                test_tar = TemporaryFile(mode='rb', suffix='.tar.gz')
                test_files = {
                    'file': ('test.tar.gz', test_tar, 'application/vnd.redhat.advisor.collection+tgz'),
                    'metadata': '{\"test\": \"test\"}'
                }
                test_req = self.post(url, files=test_files)
            elif method == "GET":
                test_req = self.get(url)
            if test_req.status_code in (200, 201, 202):
                logger.info(
                    "Successfully connected to: %s", url)
                return True
            else:
                logger.info("Connection failed")
                return False
        except requests.ConnectionError as exc:
            last_ex = exc
            logger.error(
                "Could not successfully connect to: %s", url)
            print(exc)
        if last_ex:
            raise last_ex

    def test_connection(self, rc=0):
        """
        Test connection to Red Hat
        """
        logger.debug("Proxy config: %s", self.proxies)
        try:
            logger.info("=== Begin Upload URL Connection Test ===")
            upload_success = self._test_urls(self.upload_url, "POST")
            logger.info("=== End Upload URL Connection Test: %s ===\n",
                        "SUCCESS" if upload_success else "FAILURE")
            logger.info("=== Begin API URL Connection Test ===")
            if self.config.legacy_upload:
                api_success = self._test_urls(self.base_url, "GET")
            else:
                api_success = self._test_urls(self.base_url + '/apicast-tests/ping', 'GET')
            logger.info("=== End API URL Connection Test: %s ===\n",
                        "SUCCESS" if api_success else "FAILURE")
            if upload_success and api_success:
                logger.info("Connectivity tests completed successfully")
                print("See %s for more details." % self.config.logging_file)
            else:
                logger.info("Connectivity tests completed with some errors")
                print("See %s for more details." % self.config.logging_file)
                rc = 1
        except requests.ConnectionError as exc:
            print(exc)
            logger.error('Connectivity test failed! '
                         'Please check your network configuration')
            print('Additional information may be in %s' % self.config.logging_file)
            return 1
        return rc

    def handle_fail_rcs(self, req):
        """
        Bail out if we get a 401 and leave a message
        """

        # attempt to read the HTTP response JSON message
        try:
            logger.log(NETWORK, "HTTP Response Message: %s", req.json()["message"])
        except:
            logger.debug("No HTTP Response message present.")

        # handle specific status codes
        if req.status_code >= 400:
            logger.debug("Debug Information:\nHTTP Status Code: %s",
                        req.status_code)
            logger.debug("HTTP Status Text: %s", req.reason)
            if req.status_code == 401:
                logger.error("Please ensure that the system is registered "
                             "with RHSM for CERT auth, or that correct "
                             "credentials are set in %s for BASIC auth.", self.config.conf)
                logger.log(NETWORK, "HTTP Response Text: %s", req.text)
            if req.status_code == 402:
                # failed registration because of entitlement limit hit
                logger.debug('Registration failed by 402 error.')
                try:
                    logger.error(req.json()["message"])
                except LookupError:
                    logger.error("Got 402 but no message")
                    logger.log(NETWORK, "HTTP Response Text: %s", req.text)
                except:
                    logger.error("Got 402 but no message")
                    logger.log(NETWORK, "HTTP Response Text: %s", req.text)
            if req.status_code == 403 and self.auto_config:
                # Insights disabled in satellite
                rhsm_hostname = urlparse(self.base_url).hostname
                if (rhsm_hostname != 'subscription.rhn.redhat.com' and
                   rhsm_hostname != 'subscription.rhsm.redhat.com'):
                    logger.error('Please enable Insights on Satellite server '
                                 '%s to continue.', rhsm_hostname)
            if req.status_code == 412:
                try:
                    unreg_date = req.json()["unregistered_at"]
                    logger.error(req.json()["message"])
                    write_unregistered_file(unreg_date)
                except LookupError:
                    unreg_date = "412, but no unreg_date or message"
                    logger.log(NETWORK, "HTTP Response Text: %s", req.text)
                except:
                    unreg_date = "412, but no unreg_date or message"
                    logger.log(NETWORK, "HTTP Response Text: %s", req.text)
            if req.status_code == 413:
                logger.error('Archive is too large to upload.')
            if req.status_code == 415:
                logger.error('Invalid content-type.')
            return True
        return False

    def get_satellite5_info(self, branch_info):
        """
        Get remote_leaf for Satellite 5 Managed box
        """
        logger.debug(
            "Remote branch not -1 but remote leaf is -1, must be Satellite 5")
        if os.path.isfile('/etc/sysconfig/rhn/systemid'):
            logger.debug("Found systemid file")
            sat5_conf = ET.parse('/etc/sysconfig/rhn/systemid').getroot()
            leaf_id = None
            for member in sat5_conf.getiterator('member'):
                if member.find('name').text == 'system_id':
                    logger.debug("Found member 'system_id'")
                    leaf_id = member.find('value').find(
                        'string').text.split('ID-')[1]
                    logger.debug("Found leaf id: %s", leaf_id)
                    branch_info['remote_leaf'] = leaf_id
            if leaf_id is None:
                logger.error("Could not determine leaf_id!  Exiting!")
                return False

    def get_branch_info(self):
        """
        Retrieve branch_info from Satellite Server
        """
        # branch_info = None
        # if os.path.exists(constants.cached_branch_info):
        #     # use cached branch info file if less than 5 minutes old
        #     #  (failsafe, should be deleted at end of client run normally)
        #     logger.debug(u'Reading branch info from cached file.')
        #     ctime = datetime.utcfromtimestamp(
        #         os.path.getctime(constants.cached_branch_info))
        #     if datetime.utcnow() < (ctime + timedelta(minutes=5)):
        #         with io.open(constants.cached_branch_info, encoding='utf8', mode='r') as f:
        #             branch_info = json.load(f)
        #         return branch_info
        #     else:
        #         logger.debug(u'Cached branch info is older than 5 minutes.')

        logger.debug(u'Obtaining branch information from %s',
                     self.branch_info_url)
        response = self.get(self.branch_info_url)
        if response.status_code != 200:
            logger.debug("There was an error obtaining branch information.")
            logger.debug(u'Bad status from server: %s', response.status_code)
            logger.debug("Assuming default branch information %s" % constants.default_branch_info)
            return False

        branch_info = response.json()
        logger.debug(u'Branch information: %s', json.dumps(branch_info))

        # Determine if we are connected to Satellite 5
        if ((branch_info[u'remote_branch'] != -1 and
             branch_info[u'remote_leaf'] == -1)):
            self.get_satellite5_info(branch_info)

        # logger.debug(u'Saving branch info to file.')
        # with io.open(constants.cached_branch_info, encoding='utf8', mode='w') as f:
        #     # json.dump is broke in py2 so use dumps
        #     bi_str = json.dumps(branch_info, ensure_ascii=False)
        #     f.write(bi_str)
        self.config.branch_info = branch_info
        return branch_info

    # -LEGACY-
    def create_system(self, new_machine_id=False):
        """
        Create the machine via the API
        """
        client_hostname = determine_hostname()
        machine_id = generate_machine_id(new_machine_id)

        branch_info = self.config.branch_info
        if not branch_info:
            return False

        remote_branch = branch_info['remote_branch']
        remote_leaf = branch_info['remote_leaf']

        data = {'machine_id': machine_id,
                'remote_branch': remote_branch,
                'remote_leaf': remote_leaf,
                'hostname': client_hostname}
        if self.config.display_name is not None:
            data['display_name'] = self.config.display_name
        data = json.dumps(data)
        post_system_url = self.api_url + '/v1/systems'
        logger.debug("POST System: %s", post_system_url)
        logger.debug(data)
        return self.post(post_system_url,
                         headers={'Content-Type': 'application/json'},
                         data=data)

    # -LEGACY-
    def group_systems(self, group_name, systems):
        """
        Adds an array of systems to specified group. When given group_name
        does not exist, then client tries to create a new group and adds
        given system to the new group.

        Args:
            group_name: Display name of group
            systems: Array of {'machine_id': machine_id}
        """
        system_ids = list(systems.values())
        headers = {'Content-Type': 'application/json'}
        group_path = self.api_url + '/v1/groups'

        # Description of the API is here (groups GET /groups):
        # yeah, links with anchors do not work as expected :-/
        # https://console.redhat.com/docs/api/inventory/v1#operations-groups-api
        group_get_path = group_path + ('?name=%s' % quote(group_name))
        get_group = self.get(group_get_path)

        api_group_id = None
        if get_group.status_code == 200:
            group_info = get_group.json()
            if isinstance(group_info, dict):
                try:
                    results = group_info.get("results", None)
                    if isinstance(results, list) and len(results) >= 1:
                        group = results[0]
                        if isinstance(group, dict):
                            api_group_id = group.get("id", None)
                except KeyError:
                    logger.error("Group info does not contain information about ID")

            if api_group_id is None:
                logger.warning("Unable to get ID of group: {group_name}".format(group_name=group_name))
            else:
                logger.debug("Adding systems {systems} to the group {group_name}".format(
                    systems=", ".join(system_ids), group_name=group_name
                ))
                # Documentation for this REST API is here (groups POST /groups/{group_id}/hosts):
                # yeah, links with anchors do not work as expected :-/
                # https://console.redhat.com/docs/api/inventory/v1#operations-groups-api
                data = json.dumps(system_ids)
                post_systems = self.post(
                    group_path +
                    ('/%s/hosts' % api_group_id),
                    headers=headers,
                    data=data
                )
                if post_systems.status_code == 201:
                    logger.info("Successfully added systems {systems} to the group {group_name}".format(
                        systems=", ".join(system_ids), group_name=group_name
                    ))
                else:
                    logger.error("Unable to add systems {systems} to the group {group_name}".format(
                        systems=", ".join(system_ids), group_name=group_name
                    ))

        if get_group.status_code == 404 or api_group_id is None:
            logger.debug("Adding systems {systems} to new group {group_name}".format(
                systems=", ".join(system_ids), group_name=group_name
            ))

            # Group does not exist, POST to create new group with host_ids
            # Documentation for this REST API end-point is here:
            # https://console.redhat.com/docs/api/inventory/v1#operations-groups-api
            data = json.dumps({"name": group_name, "host_ids": system_ids})
            post_group = self.post(
                group_path,
                headers=headers,
                data=data
            )
            self.handle_fail_rcs(post_group)
            if post_group.status_code == 201:
                logger.info("Successfully created group {group_name} containing {systems}".format(
                    systems=", ".join(system_ids), group_name=group_name
                ))
            else:
                logger.error("Unable to create group: {group_name}".format(group_name=group_name))

    # -LEGACY-
    # Keeping this function around because it's not private and I don't know if anything else uses it
    def do_group(self):
        """
        Do grouping on register
        """
        group_id = self.config.group
        systems = {'machine_id': generate_machine_id()}
        self.group_systems(group_id, systems)

    # -LEGACY-
    def _legacy_api_registration_check(self):
        '''
        Check registration status through API
            True    system exists in inventory
            False   connection or parsing response error
            None    system is not yet registered
            string system is unregistered
        '''
        logger.debug('Checking registration status...')
        if not machine_id_exists():
            return None
        machine_id = generate_machine_id()
        try:
            url = self.api_url + '/v1/systems/' + machine_id
            res = self.get(url)
        except requests.ConnectionError:
            # can't connect, run connection test
            logger.error('Connection timed out. Running connection test...')
            self.test_connection()
            return False
        # had to do a quick bugfix changing this around,
        #   which makes the None-False-True dichotomy seem weird
        #   TODO: reconsider what gets returned, probably this:
        #       True for registered
        #       False for unregistered
        #       None for system 404
        if res.status_code != 200:
            self.handle_fail_rcs(res)
        if res.status_code not in (200, 404):
            # Network error returns False
            return False
        else:
            try:
                # check the 'unregistered_at' key of the response
                unreg_status = json.loads(res.content).get('unregistered_at', 'undefined')
                # set the global account number
                self.config.account_number = json.loads(res.content).get('account_number', 'undefined')
            except ValueError:
                # bad response, no json object
                return False
            if unreg_status == 'undefined':
                # key not found, machine not yet registered
                return None
            elif unreg_status is None:
                # unregistered_at = null, means this machine IS registered
                return True
            else:
                # machine has been unregistered, this is a timestamp
                # This is done for legacy servers that responded with the timestamp of disconnection
                # TODO: consider to remove this condition
                return unreg_status

    def _fetch_system_by_machine_id(self):
        '''
        Get a system by machine ID
        Returns
            dict    system exists in inventory
            False   system does not exist in inventory
            None    error connection or parsing response
        '''

        if not machine_id_exists():
            return False
        machine_id = generate_machine_id()
        try:
            # [circus music]
            if self.config.legacy_upload:
                url = self.base_url + '/platform/inventory/v1/hosts?insights_id=' + machine_id
            else:
                url = self.inventory_url + '/hosts?insights_id=' + machine_id
            res = self.get(url)
        except REQUEST_FAILED_EXCEPTIONS as e:
            _api_request_failed(e)
            return None
        try:
            if (self.handle_fail_rcs(res)):
                return None
            res_json = json.loads(res.content)
        except ValueError as e:
            logger.error(e)
            logger.error('Could not parse response body.')
            return None
        if res_json['total'] == 0:
            logger.debug('No hosts found with machine ID: %s', machine_id)
            return False
        return res_json['results']

    def api_registration_check(self):
        '''
            Reach out to the inventory API to check
            whether a machine exists.

            Returns
                True    system exists in inventory
                False   system does not exist in inventory
                None    error connection or parsing response
        '''
        if self.config.legacy_upload:
            return self._legacy_api_registration_check()

        logger.debug('Checking registration status...')
        results = self._fetch_system_by_machine_id()
        if not results:
            return results

        logger.debug('System found.')
        logger.debug('Machine ID: %s', results[0]['insights_id'])
        logger.debug('Inventory ID: %s', results[0]['id'])
        return True

    # -LEGACY-
    def _legacy_unregister(self):
        """
        Unregister this system from the insights service
        """
        machine_id = generate_machine_id()
        try:
            logger.debug("Unregistering %s", machine_id)
            url = self.api_url + "/v1/systems/" + machine_id
            self.delete(url)
            logger.info(
                "Successfully unregistered from the Red Hat Insights Service")
            return True
        except requests.ConnectionError as e:
            logger.debug(e)
            logger.error("Could not unregister this system")
            return False

    def unregister(self):
        """
        Unregister this system from the insights service
        """
        if self.config.legacy_upload:
            return self._legacy_unregister()

        results = self._fetch_system_by_machine_id()
        if not results:
            logger.info('This host could not be found.')
            return False
        try:
            logger.debug("Unregistering host...")
            url = self.inventory_url + "/hosts/" + results[0]['id']
            response = self.delete(url)
            response.raise_for_status()
            logger.info(
                "Successfully unregistered from the Red Hat Insights Service")
            return True
        except (requests.ConnectionError, requests.Timeout, requests.HTTPError) as e:
            logger.debug(e)
            logger.error("Could not unregister this system")
            return False

    # -LEGACY-
    def register(self):
        """
        Register this machine
        """
        client_hostname = determine_hostname()
        # This will undo a blacklist
        logger.debug("API: Create system")
        system = self.create_system(new_machine_id=False)
        if system is False:
            return ('Could not reach the Insights service to register.', '', '', '')

        # If we get a 409, we know we need to generate a new machine-id
        if system.status_code == 409:
            system = self.create_system(new_machine_id=True)
        self.handle_fail_rcs(system)

        logger.debug("System: %s", system.json())

        message = system.headers.get("x-rh-message", "")

        # Do grouping
        if self.config.group is not None:
            self.do_group()

        # Display registration success messasge to STDOUT and logs
        if system.status_code == 201:
            try:
                system_json = system.json()
                machine_id = system_json["machine_id"]
                account_number = system_json["account_number"]
                logger.info("You successfully registered %s to account %s." % (machine_id, account_number))
            except:
                logger.debug('Received invalid JSON on system registration.')
                logger.debug('API still indicates valid registration with 201 status code.')
                logger.debug(system)
                logger.debug(system.json())

        if self.config.group is not None:
            return (message, client_hostname, self.config.group, self.config.display_name)
        elif self.config.display_name is not None:
            return (message, client_hostname, "None", self.config.display_name)
        else:
            return (message, client_hostname, "None", "")

    def _archive_too_big(self, archive_file):
        '''
        Some helpful messaging for when the archive is too large for ingress
        '''
        archive_filesize = size_in_mb(
            os.stat(archive_file).st_size)
        logger.info("Archive is {fsize} MB which is larger than the maximum allowed size of {flimit} MB.".format(
            fsize=archive_filesize, flimit=constants.archive_filesize_max))

        biggest_file = largest_spec_in_archive(archive_file)
        logger.info("The largest file in the archive is %s at %s MB.", biggest_file[0], size_in_mb(biggest_file[1]))
        logger.info("Please add the following spec to /etc/insights-client/file-redaction.yaml."
        "According to the documentation https://access.redhat.com/articles/4511681\n\n"
        "****  /etc/insights-client/file-redaction.yaml ****\n"
        "# file-redaction.yaml\n"
        "# Omit entire output of files\n"
        "# Files can be specified either by full filename or\n"
        "#   by the specs listed in insights/specs/default.py\n"
        "files:\n"
        "- %s \n**** ****", biggest_file[2])

    # -LEGACY-
    def _legacy_upload_archive(self, data_collected, duration):
        '''
        Do an HTTPS upload of the archive
        '''
        file_name = os.path.basename(data_collected)
        try:
            from insights.contrib import magic
            m = magic.open(magic.MAGIC_MIME)
            m.load()
            mime_type = m.file(data_collected)
        except ImportError:
            magic = None
            logger.debug('python-magic not installed, using backup function...')
            from .utilities import magic_plan_b
            mime_type = magic_plan_b(data_collected)

        files = {
            'file': (file_name, open(data_collected, 'rb'), mime_type)}

        upload_url = self.upload_url + '/' + generate_machine_id()

        logger.debug("Uploading %s to %s", data_collected, upload_url)

        headers = {'x-rh-collection-time': str(duration)}
        try:
            upload = self.post(upload_url, files=files, headers=headers)
        except Exception:
            raise

        if upload.status_code in (200, 201):
            the_json = json.loads(upload.text)
        else:
            logger.error("Upload archive failed with status code  %s", upload.status_code)
            if upload.status_code == 413:
                # let the user know what file is bloating the archive
                self._archive_too_big(data_collected)
            return upload
        try:
            self.config.account_number = the_json["upload"]["account_number"]
        except:
            self.config.account_number = None
        logger.debug("Upload duration: %s", upload.elapsed)
        return upload

    def upload_archive(self, data_collected, content_type, duration=None):
        """
        Do an HTTPS Upload of the archive
        """

        if self.config.legacy_upload:
            return self._legacy_upload_archive(data_collected, duration)
        file_name = os.path.basename(data_collected)
        upload_url = self.upload_url
        c_facts = {}

        try:
            c_facts = get_canonical_facts()
        except Exception as e:
            logger.debug('Error getting canonical facts: %s', e)
        if self.config.display_name:
            # add display_name to canonical facts
            c_facts['display_name'] = self.config.display_name
        if self.config.ansible_host:
            # add ansible_host to canonical facts
            c_facts['ansible_host'] = self.config.ansible_host
        if self.config.branch_info:
            c_facts["branch_info"] = self.config.branch_info
            c_facts["satellite_id"] = self.config.branch_info["remote_leaf"]
        # Clean (obfuscate and redact) the "c_facts"
        c_facts = json.dumps(self._clean_facts(c_facts))
        logger.debug('Canonical facts collected:\n%s', c_facts)

        files = {
            'file': (file_name, open(data_collected, 'rb'), content_type),
            'metadata': c_facts
        }
        logger.debug('content-type: %s', content_type)
        logger.debug("Uploading %s to %s", data_collected, upload_url)
        try:
            upload = self.post(upload_url, files=files, headers={})
        except Exception:
            raise

        logger.debug('Request ID: %s', upload.headers.get('x-rh-insights-request-id', None))
        if upload.status_code in (200, 202):
            # 202 from platform, no json response
            logger.debug(upload.text)
            # upload = registration on platform
            try:
                write_registered_file()
            except OSError as e:
                if e.errno == errno.EACCES and os.getuid() != 0:
                    # if permissions error as non-root, ignore
                    pass
                else:
                    logger.error('Could not update local registration record: %s', str(e))
        else:
            logger.debug(
                "Upload archive failed with status code %s",
                upload.status_code)
            if upload.status_code == 413:
                # let the user know what file is bloating the archive
                self._archive_too_big(data_collected)
            return upload
        logger.debug("Upload duration: %s", upload.elapsed)
        return upload

    # -LEGACY-
    def _legacy_set_display_name(self, display_name):
        machine_id = generate_machine_id()
        try:
            url = self.api_url + '/v1/systems/' + machine_id

            res = self.get(url)
            old_display_name = json.loads(res.content).get('display_name', None)
            if display_name == old_display_name:
                logger.debug('Display name unchanged: %s', old_display_name)
                return True

            res = self.put(url,
                           headers={'Content-Type': 'application/json'},
                           data=json.dumps(
                               {'display_name': display_name}))
            if res.status_code == 200:
                logger.info('System display name changed from %s to %s',
                            old_display_name,
                            display_name)
                return True
            elif res.status_code == 404:
                logger.error('System not found. '
                             'Please run insights-client --register.')
                return False
            else:
                logger.error('Unable to set display name: %s %s',
                             res.status_code, res.text)
                return False
        except REQUEST_FAILED_EXCEPTIONS + (ValueError,) as e:
            _api_request_failed(e, None)
            # can't connect, run connection test
            return False

    def set_display_name(self, display_name):
        '''
        Set display name of a system independently of upload.
        '''
        if self.config.legacy_upload:
            return self._legacy_set_display_name(display_name)

        system = self._fetch_system_by_machine_id()
        if not system:
            return system
        inventory_id = system[0]['id']

        req_url = self.inventory_url + '/hosts/' + inventory_id
        try:
            res = self.patch(req_url, json={'display_name': display_name})
        except REQUEST_FAILED_EXCEPTIONS as e:
            _api_request_failed(e)
            return False
        if (self.handle_fail_rcs(res)):
            logger.error('Could not update display name.')
            return False
        logger.info('Display name updated to ' + display_name + '.')
        return True

    def set_ansible_host(self, ansible_host):
        '''
        Set Ansible hostname of a system independently of upload.
        '''
        system = self._fetch_system_by_machine_id()
        if not system:
            return system
        inventory_id = system[0]['id']

        req_url = self.inventory_url + '/hosts/' + inventory_id
        try:
            res = self.patch(req_url, json={'ansible_host': ansible_host})
        except REQUEST_FAILED_EXCEPTIONS as e:
            _api_request_failed(e)
            return False
        if (self.handle_fail_rcs(res)):
            logger.error('Could not update Ansible hostname.')
            return False
        logger.info('Ansible hostname updated to ' + ansible_host + '.')
        return True

    def get_diagnosis(self, remediation_id=None):
        '''
            Reach out to the platform and fetch a diagnosis.
            Spirtual successor to --to-json from the old client.
        '''
        # this uses machine id as identifier instead of inventory id
        diag_url = self.base_url + '/remediations/v1/diagnosis/' + generate_machine_id()
        params = {}
        if remediation_id:
            # validate this?
            params['remediation'] = remediation_id
        try:
            res = self.get(diag_url, params=params)
        except (requests.ConnectionError, requests.Timeout) as e:
            _api_request_failed(e)
            return False
        if (self.handle_fail_rcs(res)):
            logger.error('Unable to get diagnosis data: %s %s',
                         res.status_code, res.text)
            return None
        return res.json()

    def get_advisor_report(self):
        '''
            Retrieve advisor report
        '''
        url = self.inventory_url + "/hosts?insights_id=%s" % generate_machine_id()
        res = self.get(url)
        if res.status_code not in [requests.codes.OK, requests.codes.NOT_MODIFIED]:
            return None

        host_details = json.loads(res.content)
        if host_details["total"] < 1:
            _host_not_found()
        if host_details["total"] > 1:
            raise Exception("Error: multiple hosts detected (insights_id = %s). To fix this error, run command: insights-client --unregister && insights-client --register" % generate_machine_id())

        if not os.path.exists("/var/lib/insights"):
            os.makedirs("/var/lib/insights", mode=0o755)

        with open("/var/lib/insights/host-details.json", mode="w+b") as f:
            f.write(res.content)
            logger.debug("Wrote \"/var/lib/insights/host-details.json\"")

        host_id = host_details["results"][0]["id"]
        url = self.base_url + "/insights/v1/system/%s/reports/" % host_id
        res = self.get(url)
        if res.status_code not in [requests.codes.OK, requests.codes.NOT_MODIFIED]:
            return None

        with open("/var/lib/insights/insights-details.json", mode="w+b") as f:
            f.write(res.content)
            logger.debug("Wrote \"/var/lib/insights/insights-details.json\"")

        return json.loads(res.content)

    def checkin(self):
        '''
            Sends an ultralight check-in request containing only the Canonical Facts.
        '''
        logger.info("Checking in...")

        try:
            canonical_facts = get_canonical_facts()
            canonical_facts = self._clean_facts(canonical_facts)
        except Exception as e:
            print('Error getting canonical facts: %s', e)
            logger.debug('Error getting canonical facts: %s', e)
            logger.debug('Falling back to only machine ID.')
            insights_id = generate_machine_id()
            canonical_facts = {"insights_id": str(insights_id)}

        url = self.inventory_url + "/hosts/checkin"
        logger.debug("Sending check-in request to %s with %s" % (url, canonical_facts))
        try:
            response = self.post(url, headers={"Content-Type": "application/json"}, data=json.dumps(canonical_facts))
            # Change to POST when the API is fixed.
        except REQUEST_FAILED_EXCEPTIONS as exception:
            _api_request_failed(exception)
            return None
        logger.debug("Check-in response status code %d" % response.status_code)

        if response.status_code == requests.codes.CREATED:
            # Remove OK when the API is fixed.
            logger.info("Successfully checked in!")
            return True
        elif response.status_code == requests.codes.NOT_FOUND:
            # Remove BAD_REQUEST when the API is fixed.
            _host_not_found()
        else:
            logger.debug("Check-in response body %s" % response.text)
            raise RuntimeError("Unknown check-in API response")

    def _clean_facts(self, cfacts):
        def _deep_clean(data):
            """
            Clean (obfuscate and redact) the data items one by one.
            """
            if isinstance(data, dict):
                for key, values in data.items():
                    data[key] = _deep_clean(values)
            elif isinstance(data, list):
                for i, item in enumerate(data):
                    data[i] = _deep_clean(item)
            elif isinstance(data, str):
                return cleaner.clean_content(
                    data,
                    obf_funcs=obf_funcs,
                    no_redact=False)
            return data
        # Clean (obfuscate and redact) the "c_facts"
        pc = InsightsUploadConf(self.config)
        cleaner = spec_cleaner.Cleaner(self.config, pc.get_rm_conf())
        obf_funcs = cleaner.get_obfuscate_functions()
        return _deep_clean(cfacts)



"""
Logical Volume Management configuration and status
==================================================

Parsers for lvm data based on output of various commands and file contents.

This module contains the classes that parse the output of the commands `lvs`,
`pvs`, and `vgs`, and the content of the files `/etc/lvm/lvm.conf`,
`/etc/lvm/devices/system.devices`.

Pvs - command ``/sbin/pvs --nameprefixes --noheadings --separator='|' -a -o pv_all``
------------------------------------------------------------------------------------

PvsHeadings - command ``pvs -a -v -o +pv_mda_free,pv_mda_size,pv_mda_count,pv_mda_used_count,pe_count --config="global{locking_type=0}"``
-----------------------------------------------------------------------------------------------------------------------------------------

Vgs - command ``/sbin/vgs --nameprefixes --noheadings --separator='|' -a -o vg_all``
------------------------------------------------------------------------------------

VgsHeadings - command ``vgs -v -o +vg_mda_count,vg_mda_free,vg_mda_size,vg_mda_used_count,vg_tags --config="global{locking_type=0}"``
-------------------------------------------------------------------------------------------------------------------------------------

Lvs - command ``/sbin/lvs --nameprefixes --noheadings --separator='|' -a -o lv_name,lv_size,lv_attr,mirror_log,vg_name,devices,region_size,data_percent,metadata_percent,segtype,seg_monitor --config="global{locking_type=0}"``
--------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------

LvsHeadings - command ``/sbin/lvs -a -o +lv_tags,devices --config="global{locking_type=0}"``
--------------------------------------------------------------------------------------------

LvmConf - file ``/etc/lvm/lvm.conf``
------------------------------------

LvmSystemDevices - file ``/etc/lvm/devices/system.devices``
-----------------------------------------------------------

LvmFullReport - command ``/sbin/lvm fullreport -a --reportformat json``
-----------------------------------------------------------------------

"""
from __future__ import print_function

import json

from insights.core import CommandParser, JSONParser, LegacyItemAccess, Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.parsers import get_active_lines, optlist_to_dict, parse_fixed_table
from insights.specs import Specs
from insights.util import deprecated
from insights.util import parse_keypair_lines


def map_keys(pvs, keys):
    """
    Add human readable key names to dictionary while leaving any existing key names.
    """
    rs = []
    for pv in pvs:
        r = dict((v, None) for k, v in keys.items())
        for k, v in pv.items():
            if k in keys:
                r[keys[k]] = v
            r[k] = v
        rs.append(r)
    return rs


def find_warnings(content):
    """Look for lines containing warning/error/info strings instead of data."""
    keywords = [
        k.lower()
        for k in [
            "WARNING",
            "Couldn't find device",
            "Configuration setting",
            "read failed",
            "Was device resized?",
            "Invalid argument",
            "leaked on lvs",
            "Checksum error",
            "is exported",
            "failed.",
            "Invalid metadata",
            "response failed",
            "duplicate",
            "not found",
            "Missing device",
            "Internal error",
            "Input/output error",
            "Incorrect metadata",
            "Cannot process volume",
            "No such file or directory",
            "Logging initialised",
            "changed sizes",
            "vsnprintf failed",
            "write failed",
            "correction failed",
            "Failed to write",
            "Couldn't read",
            "marked missing",
            "Attempt to close device",
            "Ignoring supplied major",
            "not match metadata",
            "Reading VG",
            "Error reading device"
        ]
    ]
    for l in content:
        lower = l.strip().lower()
        # Avoid hitting keywords inside the data
        if not lower.startswith("lvm2"):
            if any(k in lower for k in keywords):
                yield l


class LvmHeadings(CommandParser):
    """Base class for parsing LVM data in table format."""

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]


class Lvm(CommandParser):
    """Base class for parsing LVM data in key=value format."""

    def parse_content(self, content):
        if "Unrecognised field:" in content[-1]:
            raise ParseException(content[-1])
        d = {"warnings": set(find_warnings(content))}
        content = [l for l in content
                   if l not in d["warnings"] and not l.startswith("File descriptor ")]
        d["content"] = list(map_keys(parse_keypair_lines(content), self.KEYS))
        self.data = d if d else None

    def __iter__(self):
        return iter(self.data["content"])

    def __len__(self):
        return len(self.data["content"])

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.data["content"][key]
        for i in self.data["content"]:
            if i[self.PRIMARY_KEY] == key:
                return i
        return None

    @property
    def locking_disabled(self):
        """bool: Returns True if any lines in input data indicate locking is disabled."""
        return any(l for l in self.data["warnings"] if "Locking disabled" in l)

    @property
    def warnings(self):
        """list: Returns a list of lines from input data containing
            warning/error/info strings.
        """
        return self.data["warnings"]


@parser(Specs.pvs_noheadings)
class Pvs(Lvm):
    """
    Parse the output of the `/sbin/pvs --nameprefixes --noheadings --separator='|' -a -o pv_all` command.

    Parse each line in the output of pvs based on the of pvs datasource in
    `insights/specs/` Output sample of pvs::

        LVM2_PV_FMT=''|LVM2_PV_UUID=''|LVM2_DEV_SIZE='500.00m'|...
        LVM2_PV_FMT='lvm2'|LVM2_PV_UUID='JvSULk-ileq-JbuS-GGgg-jkif-thuW-zvFBEl'|LVM2_DEV_SIZE='476.45g'|...

    Returns a list like::

        [
            {
                'LVM2_PV_FMT'    : '',
                'LVM2_PV_UUID'    : '',
                'LVM2_DEV_SIZE'   : '500.00m',
                ...
            },
            {
                'LVM2_PV_FMT'    : 'lvm2',
                'LVM2_PV_UUID'    : 'JvSULk-ileq-JbuS-GGgg-jkif-thuW-zvFBEl',
                'LVM2_DEV_SIZE'   : '476.45g',
                ...
            }
        ]

    Since it is possible to have two PV's with the same name (for example *unknown device*) a
    unique key for each PV is created by joining the `PV_NAME and PV_UUID fields with a `+`
    character.  This key is added to the dictionary as the `PV_KEY` field.
    """

    KEYS = {
        "LVM2_PV_MDA_USED_COUNT": "#PMdaUse",
        "LVM2_PV_UUID": "PV_UUID",
        "LVM2_DEV_SIZE": "DevSize",
        "LVM2_PV_FMT": "Fmt",
        "LVM2_PV_MDA_FREE": "PMdaFree",
        "LVM2_PV_EXPORTED": "Exported",
        "LVM2_PV_SIZE": "PSize",
        "LVM2_PV_BA_START": "BA_start",
        "LVM2_PV_PE_ALLOC_COUNT": "Alloc",
        "LVM2_VG_NAME": "VG",
        "LVM2_PV_TAGS": "PV_Tags",
        "LVM2_PV_PE_COUNT": "PE",
        "LVM2_PV_BA_SIZE": "BA_size",
        "LVM2_PV_ATTR": "Attr",
        "LVM2_PE_START": "1st_PE",
        "LVM2_PV_USED": "Used",
        "LVM2_PV_NAME": "PV",
        "LVM2_PV_MDA_COUNT": "#PMda",
        "LVM2_PV_FREE": "PFree",
        "LVM2_PV_ALLOCATABLE": "Allocatable",
        "LVM2_PV_MDA_SIZE": "PMdaSize",
        "LVM2_PV_MISSING": "Missing",
    }

    PRIMARY_KEY = "PV"

    def parse_content(self, content):
        super(Pvs, self).parse_content(content)
        for pv in self.data["content"]:
            pv_name = pv.get("PV") if pv.get("PV") is not None else "no_name"
            pv_uuid = pv.get("PV_UUID") if pv.get("PV_UUID") is not None else "no_uuid"
            pv.update({"PV_KEY": "+".join([pv_name, pv_uuid])})

    def vg(self, name):
        """Return all physical volumes assigned to the given volume group"""
        return [i for i in self.data["content"] if i["VG"] == name]


@parser(Specs.pvs_noheadings_all)
class PvsAll(Pvs):
    """
    Parse the output of the `/sbin/pvs --nameprefixes --noheadings --separator='|' -a -o pv_all,vg_name --config='global{locking_type=0} devices{filter=["a|.*|"]}'` command.

    Uses the ``Pvs`` class defined in this module.
    """

    pass


@parser(Specs.pvs_headings)
class PvsHeadings(LvmHeadings):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.lvm.Pvs` instead.

    Parses the output of the
    `pvs -a -v -o +pv_mda_free,pv_mda_size,pv_mda_count,pv_mda_used_count,pe_count --config="global{locking_type=0}"`
    command.

    Since it is possible to have two PV's with the same name (for example *unknown device*) a
    unique key for each PV is created by joining the `PV_NAME and PV_UUID fields with a `+`
    character.  This key is added to the resulting dictionary as the `PV_KEY` field.

    Sample input::

          WARNING: Locking disabled. Be careful! This could corrupt your metadata.
            Scanning all devices to update lvmetad.
            No PV label found on /dev/loop0.
            No PV label found on /dev/loop1.
            No PV label found on /dev/sda1.
            No PV label found on /dev/fedora/root.
            No PV label found on /dev/sda2.
            No PV label found on /dev/fedora/swap.
            No PV label found on /dev/fedora/home.
            No PV label found on /dev/mapper/docker-253:1-2361272-pool.
            Wiping internal VG cache
            Wiping cache of LVM-capable devices
          PV                                                    VG     Fmt  Attr PSize   PFree DevSize PV UUID                                PMdaFree  PMdaSize  #PMda #PMdaUse PE
          /dev/fedora/home                                                  ---       0     0  418.75g                                               0         0      0        0      0
          /dev/fedora/root                                                  ---       0     0   50.00g                                               0         0      0        0      0
          /dev/fedora/swap                                                  ---       0     0    7.69g                                               0         0      0        0      0
          /dev/loop0                                                        ---       0     0  100.00g                                               0         0      0        0      0
          /dev/loop1                                                        ---       0     0    2.00g                                               0         0      0        0      0
          /dev/mapper/docker-253:1-2361272-pool                             ---       0     0  100.00g                                               0         0      0        0      0
          /dev/mapper/luks-7430952e-7101-4716-9b46-786ce4684f8d fedora lvm2 a--  476.45g 4.00m 476.45g FPLCRf-d918-LVL7-6e3d-n3ED-aiZv-EesuzY        0   1020.00k     1        1 121970
          /dev/sda1                                                         ---       0     0  500.00m                                               0         0      0        0      0
          /dev/sda2                                                         ---       0     0  476.45g                                               0         0      0        0      0
            Reloading config files
            Wiping internal VG cache

    Attributes:
        data (list): List of dicts, each dict containing one row of the table
            with column headings as keys.
        warnings (set): Set of lines from input data containing
            warning strings.

    Examples:
        >>> pvs_data[0]['PV']
        '/dev/fedora/home'
        >>> pvs_data[0]['PMdaSize']
        '0'

    """
    def __init__(self, *args, **kwargs):
        deprecated(
            PvsHeadings,
            "Please use insights.parsers.lvm.Pvs instead.",
            "3.6.0"
        )
        super(PvsHeadings, self).__init__(*args, **kwargs)

    PRIMARY_KEY = Pvs.PRIMARY_KEY

    def parse_content(self, content):
        self.warnings = set(find_warnings(content))
        content = [l for l in content if l not in self.warnings]
        self.data = parse_fixed_table(
            content,
            heading_ignore=["PV "],
            header_substitute=[("PV UUID", "PV_UUID"), ("1st PE", "1st_PE")],
            trailing_ignore=["Reloading", "Wiping"],
        )
        self.data = map_keys(self.data, Pvs.KEYS)
        for pv in self.data:
            pv_name = pv.get("PV") if pv.get("PV") is not None else "no_name"
            pv_uuid = pv.get("PV_UUID") if pv.get("PV_UUID") is not None else "no_uuid"
            pv.update({"PV_KEY": "+".join([pv_name, pv_uuid])})

    def vg(self, name):
        """Return all physical volumes assigned to the given volume group"""
        return [i for i in self.data if i["VG"] == name]


@parser(Specs.vgs_noheadings)
class Vgs(Lvm):
    """
    Parse the output of the `/sbin/vgs --nameprefixes --noheadings --separator='|' -a -o vg_all` command.

    Parse each line in the output of vgs based on the vgs datasource in
    `insights/specs/` Output sample of vgs::

        LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='YCpusB-LEly-THGL-YXhC-t3q6-mUQV-wyFZrx'|LVM2_VG_NAME='rhel'|LVM2_VG_ATTR='wz--n-'|...
        LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='123456-LEly-THGL-YXhC-t3q6-mUQV-123456'|LVM2_VG_NAME='fedora'|LVM2_VG_ATTR='wz--n-'|...

    Returns a list like::

        [
            {
                'LVM2_PV_FMT'    : 'lvm2',
                'LVM2_VG_UUID'    : 'YCpusB-LEly-THGL-YXhC-t3q6-mUQV-wyFZrx',
                'LVM2_VG_NAME'   : 'rhel',
                ...
            },
            {
                'LVM2_PV_FMT'    : 'lvm2',
                'LVM2_VG_UUID'    : '123456-LEly-THGL-YXhC-t3q6-mUQV-123456',
                'LVM2_VG_NAME'   : 'fedora',
                ...
            }
        ]
    """

    KEYS = {
        "LVM2_VG_EXTENDABLE": "Extendable",
        "LVM2_VG_EXTENT_SIZE": "Ext",
        "LVM2_VG_MDA_COUNT": "#VMda",
        "LVM2_VG_PROFILE": "VProfile",
        "LVM2_VG_ALLOCATION_POLICY": "AllocPol",
        "LVM2_MAX_PV": "MaxPV",
        "LVM2_VG_UUID": "VG_UUID",
        "LVM2_VG_ATTR": "Attr",
        "LVM2_VG_SYSID": "SYS_ID",
        "LVM2_VG_MDA_USED_COUNT": "#VMdaUse",
        "LVM2_VG_MDA_FREE": "VMdaFree",
        "LVM2_VG_LOCKTYPE": "Lock_Type",
        "LVM2_VG_LOCK_TYPE": "Lock_Type",
        "LVM2_VG_TAGS": "VG_Tags",
        "LVM2_VG_FMT": "Fmt",
        "LVM2_PV_COUNT": "#PV",
        "LVM2_VG_EXTENT_COUNT": "#Ext",
        "LVM2_VG_MDA_SIZE": "VMdaSize",
        "LVM2_SNAP_COUNT": "#SN",
        "LVM2_VG_EXPORTED": "Exported",
        "LVM2_LV_COUNT": "#LV",
        "LVM2_VG_NAME": "VG",
        "LVM2_VG_MDA_COPIES": "#VMdaCps",
        "LVM2_VG_SYSTEMID": "System_ID",
        "LVM2_VG_FREE": "VFree",
        "LVM2_VG_SEQNO": "Seq",
        "LVM2_VG_FREE_COUNT": "Free",
        "LVM2_VG_PARTIAL": "Partial",
        "LVM2_VG_PERMISSIONS": "VPerms",
        "LVM2_VG_CLUSTERED": "Clustered",
        "LVM2_VG_LOCKARGS": "Lock Args",
        "LVM2_VG_LOCK_ARGS": "Lock Args",
        "LVM2_MAX_LV": "MaxLV",
        "LVM2_VG_SIZE": "VSize",
    }

    PRIMARY_KEY = "VG"


@parser(Specs.vgs_noheadings_all)
class VgsAll(Vgs):
    """
    Parse the output of the `/sbin/vgs --nameprefixes --noheadings --separator='|' -a -o vg_all --config='global{locking_type=0} devices{filter=[\"a|.*|\"]}'` command.

    Uses the ``Vgs`` class defined in this module.
    """

    pass


@parser(Specs.vgs_headings)
class VgsHeadings(LvmHeadings):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.lvm.Vgs` instead.

    Parses output of the
    `vgs -v -o +vg_mda_count,vg_mda_free,vg_mda_size,vg_mda_used_count,vg_tags --config="global{locking_type=0}"` command.

    Sample input::

          WARNING: Locking disabled. Be careful! This could corrupt your metadata.
            Using volume group(s) on command line.
          VG            Attr   Ext   #PV #LV #SN VSize   VFree    VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
          DATA_OTM_VG   wz--n- 4.00m   6   1   0   2.05t 1020.00m xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc              6   507.00k  1020.00k        6
          ITM_VG        wz--n- 4.00m   1   1   0  16.00g    4.00m nws5dd-INe6-1db6-9U1N-F0G3-S1z2-5XTdO4              1   508.00k  1020.00k        1
          ORABIN_OTM_VG wz--n- 4.00m   2   3   0 190.00g       0  hfJwg8-hset-YgUY-X6NJ-gkWE-EunZ-KuCXGP              2   507.50k  1020.00k        2
          REDO_OTM_VG   wz--n- 4.00m   1   3   0  50.00g       0  Q2YtGy-CWKU-sEYj-mqHk-rbdP-Hzup-wi8jsf              1   507.50k  1020.00k        1
          SWAP_OTM_VG   wz--n- 4.00m   1   1   0  24.00g    8.00g hAerzZ-U8QU-ICkc-xxCj-N2Ny-rWzq-pmTpWJ              1   508.00k  1020.00k        1
          rootvg        wz--n- 4.00m   1   6   0  19.51g    1.95g p4tLLb-ikeo-Ankk-2xJ6-iHYf-D4E6-KFCFvr              1   506.50k  1020.00k        1
            Reloading config files
            Wiping internal VG cache

    Attributes:
        data (list): List of dicts, each dict containing one row of the table
            with column headings as keys.
        warnings (set): Set of lines from input data containing
            warning strings.

    Examples:
        >>> vgs_info.data[0]['VG']
        'DATA_OTM_VG'
        >>> vgs_info.data[0]['VG_UUID']
        'xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc'
    """
    def __init__(self, *args, **kwargs):
        deprecated(
            VgsHeadings,
            "Please use insights.parsers.lvm.Vgs instead.",
            "3.6.0"
        )
        super(VgsHeadings, self).__init__(*args, **kwargs)

    PRIMARY_KEY = Vgs.PRIMARY_KEY

    def parse_content(self, content):
        self.warnings = set(find_warnings(content))
        content = [l for l in content if l not in self.warnings]
        self.data = parse_fixed_table(
            content,
            heading_ignore=["VG "],
            header_substitute=[("VG Tags", "VG_Tags"), ("VG UUID", "VG_UUID")],
            trailing_ignore=["Reloading", "Wiping"],
        )
        self.data = map_keys(self.data, Vgs.KEYS)


@parser(Specs.lvs_noheadings)
class Lvs(Lvm):
    """
    Parse the output of the `/sbin/lvs --nameprefixes --noheadings --separator='|' -a -o lv_name,lv_size,lv_attr,mirror_log,vg_name,devices,region_size,data_percent,metadata_percent,segtype,seg_monitor --config="global{locking_type=0}"` command.

    Parse each line in the output of lvs based on the lvs datasource in
    `insights/specs/`:

    Output sample of lvs::

        LVM2_LV_UUID='KX68JI-8ISN-YedH-ZYDf-yZbK-zkqE-3aVo6m'|LVM2_LV_NAME='docker-poolmeta'|LVM2_LV_FULL_NAME='rhel/docker-poolmeta'|...
        LVM2_LV_UUID='123456-8ISN-YedH-ZYDf-yZbK-zkqE-123456'|LVM2_LV_NAME='rhel_root'|LVM2_LV_FULL_NAME='rhel/rhel_root'|LVM2_LV_PATH='/dev/rhel/docker-poolmeta'|...

    Return a list, as shown below::

        [
            {
                'LVM2_LV_UUID'      : 'KX68JI-8ISN-YedH-ZYDf-yZbK-zkqE-3aVo6m',
                'LVM2_LV_NAME'      : 'docker-poolmeta',
                'LVM2_LV_FULL_NAME'   : 'rhel/docker-poolmeta',
                ...
            },
            {
                'LVM2_LV_UUID'      : '123456-8ISN-YedH-ZYDf-yZbK-zkqE-123456',
                'LVM2_LV_NAME'      : 'rhel_root',
                'LVM2_LV_FULL_NAME'   : 'rhel/rhel_root',
                ...
            }
        ]
    """

    KEYS = {
        "LVM2_POOL_LV_UUID": "Pool_UUID",
        "LVM2_LV_PARENT": "Parent",
        "LVM2_LV_SKIP_ACTIVATION": "SkipAct",
        "LVM2_LV_HEALTH_STATUS": "Health",
        "LVM2_LV_KERNEL_MINOR": "KMin",
        "LVM2_RAID_WRITE_BEHIND": "WBehind",
        "LVM2_LV_ANCESTORS": "Ancestors",
        "LVM2_LV_TIME": "Time",
        "LVM2_METADATA_PERCENT": "Meta%",
        "LVM2_LV_DM_PATH": "DMPath",
        "LVM2_LV_INACTIVE_TABLE": "InactiveTable",
        "LVM2_LV_UUID": "LV_UUID",
        "LVM2_LV_MODULES": "Modules",
        "LVM2_DEVICES": "Devices",
        "LVM2_LV_ACTIVE_REMOTELY": "ActRemote",
        "LVM2_LV_ACTIVE_LOCALLY": "ActLocal",
        "LVM2_LV_TAGS": "LV_Tags",
        "LVM2_LV_IMAGE_SYNCED": "ImgSynced",
        "LVM2_CACHE_WRITE_MISSES": "CacheWriteMisses",
        "LVM2_LV_PERMISSIONS": "LPerms",
        "LVM2_CACHE_TOTAL_BLOCKS": "CacheTotalBlocks",
        "LVM2_LV_ACTIVE_EXCLUSIVELY": "ActExcl",
        "LVM2_LV_PATH": "Path",
        "LVM2_LV_FULL_NAME": "LV",
        "LVM2_LV_READ_AHEAD": "Rahead",
        "LVM2_SNAP_PERCENT": "Snap%",
        "LVM2_CACHE_WRITE_HITS": "CacheWriteHits",
        "LVM2_MIRROR_LOG": "Log",
        "LVM2_CACHE_DIRTY_BLOCKS": "CacheDirtyBlocks",
        "LVM2_SEG_COUNT": "#Seg",
        "LVM2_MOVE_PV": "Move",
        "LVM2_LV_FIXED_MINOR": "FixMin",
        "LVM2_SYNC_PERCENT": "Cpy%Sync",
        "LVM2_LV_METADATA_SIZE": "MSize",
        "LVM2_LV_ATTR": "Attr",
        "LVM2_RAID_MAX_RECOVERY_RATE": "MaxSync",
        "LVM2_LV_DEVICE_OPEN": "DevOpen",
        "LVM2_LV_ALLOCATION_POLICY": "AllocPol",
        "LVM2_LV_MERGING": "Merging",
        "LVM2_LV_SIZE": "LSize",
        "LVM2_LV_MAJOR": "Maj",
        "LVM2_ORIGIN_SIZE": "OSize",
        "LVM2_RAID_SYNC_ACTION": "SyncAction",
        "LVM2_MIRROR_LOG_UUID": "Log_UUID",
        "LVM2_POOL_LV": "Pool",
        "LVM2_COPY_PERCENT": "Cpy%Sync",
        "LVM2_CONVERT_LV": "Convert",
        "LVM2_LV_KERNEL_READ_AHEAD": "KRahead",
        "LVM2_LV_NAME": "LV",
        "LVM2_LV_HOST": "Host",
        "LVM2_CACHE_USED_BLOCKS": "CacheUsedBlocks",
        "LVM2_RAID_MIN_RECOVERY_RATE": "MinSync",
        "LVM2_ORIGIN_UUID": "Origin_UUID",
        "LVM2_LV_SUSPENDED": "Suspended",
        "LVM2_RAID_MISMATCH_COUNT": "Mismatches",
        "LVM2_LV_KERNEL_MAJOR": "KMaj",
        "LVM2_LV_LAYOUT": "Layout",
        "LVM2_LV_PROFILE": "LProfile",
        "LVM2_LV_LIVE_TABLE": "LiveTable",
        "LVM2_LV_INITIAL_IMAGE_SYNC": "InitImgSync",
        "LVM2_LV_CONVERTING": "Converting",
        "LVM2_CACHE_READ_HITS": "CacheReadHits",
        "LVM2_VG_NAME": "VG",
        "LVM2_METADATA_LV": "Meta",
        "LVM2_LV_ACTIVE": "Active",
        "LVM2_CONVERT_LV_UUID": "Convert_UUID",
        "LVM2_LV_MERGE_FAILED": "MergeFailed",
        "LVM2_METADATA_LV_UUID": "Meta_UUID",
        "LVM2_LV_ROLE": "Role",
        "LVM2_LV_WHEN_FULL": "WhenFull",
        "LVM2_LV_ALLOCATION_LOCKED": "AllocLock",
        "LVM2_DATA_PERCENT": "Data%",
        "LVM2_LV_LOCKARGS": "Lock_Args",
        "LVM2_LV_SNAPSHOT_INVALID": "SnapInvalid",
        "LVM2_MOVE_PV_UUID": "Move_UUID",
        "LVM2_LV_MINOR": "Min",
        "LVM2_ORIGIN": "Origin",
        "LVM2_DATA_LV_UUID": "Data_UUID",
        "LVM2_DATA_LV": "Data",
        "LVM2_CACHE_READ_MISSES": "CacheReadMisses",
        "LVM2_LV_DESCENDANTS": "Descendants",
        "LVM2_REGION_SIZE": "Region",
        "LVM2_SEGTYPE": "SegType",
        "LVM2_SEG_MONITOR": "Monitor",
    }

    PRIMARY_KEY = "LV"

    def parse_content(self, content):
        super(Lvs, self).parse_content(content)
        for item in self.data["content"]:
            lv_name = item["LV"]
            if "/" in lv_name:
                # Reduce full name to just the name
                # This is due to the lvs command having *two identical keys*
                # with different values
                item["LV"] = lv_name.split("/")[1]

    def vg(self, name):
        """Return all logical volumes in the given volume group"""
        return [i for i in self.data["content"] if i["VG"] == name]


@parser(Specs.lvs_noheadings_all)
class LvsAll(Lvs):
    """
    Parse the output of the `/sbin/lvs --nameprefixes --noheadings --separator='|' -a -o lv_name,lv_size,lv_attr,mirror_log,vg_name,devices,region_size,data_percent,metadata_percent --config='global{locking_type=0} devices{filter=["a|.*|"]}'` command.

    Uses the ``Lvs`` class defined in this module.
    """

    pass


@parser(Specs.lvs_headings)
class LvsHeadings(LvmHeadings):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.lvm.Lvs` instead.

    Process output of the command `/sbin/lvs -a -o +lv_tags,devices --config="global{locking_type=0}"`.

    Sample Input data::

        WARNING: Locking disabled. Be careful! This could corrupt your metadata.
        LV          VG      Attr       LSize  Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert LV Tags Devices
        lv_app      vg_root -wi-ao---- 71.63g                                                             /dev/sda2(7136)
        lv_home     vg_root -wi-ao----  2.00g                                                             /dev/sda2(2272)
        lv_opt      vg_root -wi-ao----  5.00g                                                             /dev/sda2(2784)
        lv_root     vg_root -wi-ao----  5.00g                                                             /dev/sda2(0)
        lv_tmp      vg_root -wi-ao----  1.00g                                                             /dev/sda2(4064)
        lv_usr      vg_root -wi-ao----  5.00g                                                             /dev/sda2(4320)
        lv_usrlocal vg_root -wi-ao----  1.00g                                                             /dev/sda2(5600)
        lv_var      vg_root -wi-ao----  5.00g                                                             /dev/sda2(5856)
        swap        vg_root -wi-ao----  3.88g                                                             /dev/sda2(1280)

    Attributes:
        data (list): List of dicts, each dict containing one row of the table
            with column headings as keys.
        warnings (set): Set of lines from input data containing
            warning strings.

    Examples:
        >>> lvs_info.data[0]['Devices']
        '/dev/sda2(7136)'
        >>> lvs_info.data[1]['LSize']
        '2.00g'
    """
    def __init__(self, *args, **kwargs):
        deprecated(
            LvsHeadings,
            "Please use insights.parsers.lvm.Lvs instead.",
            "3.6.0"
        )
        super(LvsHeadings, self).__init__(*args, **kwargs)

    PRIMARY_KEY = Lvs.PRIMARY_KEY

    def parse_content(self, content):
        self.warnings = set(find_warnings(content))
        content = [l for l in content if l not in self.warnings]
        self.data = parse_fixed_table(
            content, heading_ignore=["LV "], header_substitute=[("LV Tags", "LV_Tags")]
        )
        self.data = map_keys(self.data, Lvs.KEYS)


KEYS_WITH_SPACES = []
for cls in (Lvs, Pvs, Vgs):
    KEYS_WITH_SPACES.extend([k for k in cls.KEYS.values() if " " in k])


LVM_CONF_FILTERS = [
    "locking_type",  # CMIRROR_PERF_ISSUE
    "filter",  # LVM_CONF_REMOVE_BOOTDEV HA_LVM_RELOCATE_ISSUE LVM_FILTER_ISSUE
    "volume_list",  # HA_LVM_RELOCATE_ISSUE
]
add_filter(Specs.lvm_conf, LVM_CONF_FILTERS)


@parser(Specs.lvm_conf)
class LvmConf(LegacyItemAccess, Parser):
    """
    Parses contents of the `/etc/lvm/lvm.conf` file.

    Sample Input::

        locking_type = 1
        #locking_type = 2
        # volume_list = [ "vg1", "vg2/lvol1", "@tag1", "@*" ]
        volume_list = [ "vg2", "vg3/lvol3", "@tag2", "@*" ]
        # filter = [ "a|loop|", "r|/dev/hdc|", "a|/dev/ide|", "r|.*|" ]

        filter = [ "r/sda[0-9]*$/",  "a/sd.*/" ]
        filter = [ "a/sda[0-9]*$/",  "r/sd.*/" ]
        shell {
            history_size = 100
        }

    Examples:
        >>> 'vg2' in lvm_conf_data.data.get('volume_list')
        True
        >>> lvm_conf_data.get("locking_type")
        1
    """

    def parse_content(self, content):
        """
        Returns a dict:
        locking_type : 1
        filter : ['a/sda[0-9]*$/', 'r/sd.*/']
        volume_list : ['vg2', 'vg3/lvol3', '@tag2', '@*']
        """
        lvm_conf_dict = {}
        for line in get_active_lines(content):
            if "=" in line:
                (key, value) = [item.strip() for item in line.split("=", 1)]
                try:
                    lvm_conf_dict[key] = json.loads(value)
                except Exception:
                    lvm_conf_dict[key] = value
        self.data = lvm_conf_dict


@parser(Specs.lvmconfig)
class LvmConfig(CommandParser):
    def parse_content(self, content):
        self.data = dict()
        key = None
        for line in content:
            line = line.rstrip()
            if not line:
                continue

            if line[-1] == "{":
                key = line.split()[0]
            elif line[0] == "}":
                key = None
            elif line[0] == "\t":
                k, v = line.strip().split("=", 1)
                # umask=077 will raise exception, and also no need to
                # transfer it, just keep it as string
                if k != 'umask':
                    try:
                        v = json.loads(v)
                    except Exception:
                        raise ParseException("Failed to parse line %s." % line)
                self.data.setdefault(key, {}).update({k: v})
            else:
                pass  # inferring this a stderr, so skipping


@parser(Specs.lvm_system_devices)
class LvmSystemDevices(Parser, dict):
    """
    Parse the content of the ``/etc/lvm/devices/system.devices`` file.
    It returns a dict. The key is the device id, the value is a dict of
    other info.

    Sample input::

        VERSION=1.1.2
        IDTYPE=devname IDNAME=/dev/vda2 DEVNAME=/dev/vda2 PVID=phl0clFbAokp9UXqbIgI5YYQxuTIJVkD PART=2

    Sample output::

        {
            '/dev/vda2': {
                'IDTYPE': 'devname',
                'DEVNAME': '/dev/vda2',
                'PVID': 'phl0clFbAokp9UXqbIgI5YYQxuTIJVkD',
                'PART': '2'
            }
        }

    Example:
        >>> type(devices)
        <class 'insights.parsers.lvm.LvmSystemDevices'>
        >>> devices['/dev/vda2']['IDTYPE']
        'devname'
        >>> devices['/dev/vda2']['PVID']
        'phl0clFbAokp9UXqbIgI5YYQxuTIJVkD'

    Raises:
        SkipComponent: when there is no device info.
    """

    def parse_content(self, content):
        for line in content:
            if 'IDNAME' in line:
                dict_info = optlist_to_dict(line, opt_sep=None)
                self[dict_info.pop('IDNAME')] = dict_info
        if not self:
            raise SkipComponent("No valid content.")


@parser(Specs.lvm_fullreport)
class LvmFullReport(JSONParser):
    """
    Parse the output of the command ``/usr/sbin/lvm fullreport -a --reportformat json``.
    Output is in JSON format so the JSONParser will be used.

    Sample input (actual input is in JSON format, data has been changed here to show the relationship
    between volume groups, physical valumes and logical volumes)::

        {
            "report": [
                {
                    "vg": [ { vg_name:vg1, vg_uuid:111, properties of vg1 } ],
                    "pv": [ { pv_name:/dev/foo, pv_uuid:222, properties of foo },
                            { pv_name:/dev/bar, pv_uuid:333, properties of bar },
                            more pvs ],
                    "lv": [ { lv_name:lv1, lv_uuid:444, properties of lv1 },
                            { lv_name:lv2, lv_uuid:555, properties of lv2 },
                            more lvs ],
                    "pvseg": [ { ..., pv_uuid:222, ... },
                               { ..., pv_uuid:222, ... },
                               { ..., pv_uuid:333, ... },
                               more pvsegs ],
                    "seg": [ { ..., lv_uuid:444, ... },
                             { ..., lv_uuid:555, ... },
                             { ..., lv_uuid:555, ... },
                             more segs ]
                },
                {
                    "vg": [ { vg_name:rhel, vg_uuid:112, properties of rhel } ],
                    "pv": [ { pv_name:/dev/foo2, pv_uuid:223, properties of foo2 },
                            { pv_name:/dev/bar2, pv_uuid:334, properties of bar2 },
                            more pvs ],
                    "lv": [ { lv_name:lv12, lv_uuid:442, properties of lv12 },
                            { lv_name:lv22, lv_uuid:5552, properties of lv22 },
                            more lvs ],
                    "pvseg": [ { ..., pv_uuid:223, ... },
                               { ..., pv_uuid:223, ... },
                               { ..., pv_uuid:334, ... },
                               more pvsegs ],
                    "seg": [ { ..., lv_uuid:442, ... },
                             { ..., lv_uuid:5552, ... },
                             { ..., lv_uuid:5552, ... },
                             more segs ]
                },
                # Orphan volume group (pv's not associated with a vg)
                {
                    "pv": [ {...} ],
                    "pvseg": [ {...} ]
                }
            ]
        }

    Output will be a python object in the same structure as the JSON::

        {
            "report": [
                {
                  "vg": [ {"vg_name": "vg1", ...}, ... ],
                  "pv": [ {"pv_name": "/dev/sdg", ...), ... ],
                  "lv": [ {"lv_name": "[fast1_cvol]", ...}, ... ],
                  "pvseg": [ {"pvseg_start": "0", ...}, ... ],
                  "seg": [ {"segtype": "cache-pool", ...}, ... ]
                },
                {
                  "vg": [ {"vg_name": "rhel", ...}, ... ],
                  "pv": [ {"pv_name": "/dev/sdd", ...), ... ],
                  "lv": [ {"lv_name": "[lvm_lock]", ...}, ... ],
                  "pvseg": [ {"pvseg_start": "0", ...}, ... ],
                  "seg": [ {"segtype": "linear", ...}, ... ]
                },
                ...
                {
                  "pv": [ {...} ],
                  "pvseg": [ {...} ]
                }
            ]
        }

    Attributes:
        warnings(list): List of warning lines included at the beginning of input
        volume_groups(dict): Dictionary with vg_name as the key, contains a dictionary including
            all information for the volume group. The vg element is represented as a list but
            will only have a single item which is that volumn group::

                {
                    "vg1": {
                        "vg": [ {"vg_name": "global", ...}, ... ],
                        "pv": [ {"pv_name": "/dev/sdd", ...), ... ],
                        "lv": [ {"lv_name": "[lvm_lock]", ...}, ... ],
                        "pvseg": [ {"pvseg_start": "0", ...}, ... ],
                        "seg": [ {"segtype": "linear", ...}, ... ]
                    },
                    ...
                }

        orphan_volume_group(dict): Dictionary of pv's and pvseg's not associated with a volume group::

            {
                "pv": [ {"pv_name": "/dev/sda", ...), ... ],
                "pvseg": [ {"pvseg_start": "0", ...}, ... ],
            }

    Example:
        >>> type(lvm_fullreport)
        <class 'insights.parsers.lvm.LvmFullReport'>
        >>> len(lvm_fullreport.volume_groups)
        2
        >>> sorted(lvm_fullreport.volume_groups.keys()) == ['rhel', 'vg1']
        True
        >>> lvm_fullreport.volume_groups['vg1']['pv'][0]['pv_name'] == '/dev/vdb'
        True

    Raises:
        SkipComponent: when there is no device info.
    """
    def parse_content(self, content):
        self.warnings = []
        skip_to_line = len(content)
        for ndx, line in enumerate(content):
            if line.strip().startswith('{'):
                skip_to_line = ndx
                break
            self.warnings.append(line)

        if skip_to_line >= len(content):
            raise SkipComponent("No LVM information in fullreport")

        super(LvmFullReport, self).parse_content(content[skip_to_line:])

        if not self.data['report']:
            raise SkipComponent("No LVM information in fullreport")

        vgs = self.data['report']
        self.volume_groups = dict()
        self.orphan_volume_group = dict()
        for vg_data in vgs:
            if 'vg' in vg_data:
                self.volume_groups[vg_data['vg'][0]['vg_name']] = vg_data
            else:
                self.orphan_volume_group = vg_data


if __name__ == "__main__":
    # This is a quick script to generate the key mappings in each subclass.
    # Run each lvm command with --separator="|", --nameprefixes and *not* --noheadings

    import sys
    from collections import OrderedDict

    content = sys.stdin.read().splitlines()
    headers = [h.strip().replace(" ", "_") for h in content[0].split("|")]
    nameprefixes = [
        v.split("=")[0].strip() for v in content[1].replace("0 ", "0").split("|")
    ]
    pairs = zip(nameprefixes, headers)
    print(json.dumps(OrderedDict(sorted(pairs))))



import doctest
import pytest

from insights.core.exceptions import ParseException, SkipComponent
from insights.parsers import proc_keyusers
from insights.parsers.proc_keyusers import ProcKeyUsers
from insights.tests import context_wrap

PROC_KEYUSERS = """
    0:   106 105/105 95/1000000 1909/25000000
  862:     4 4/4 4/200 40/20000
  980:     9 9/9 9/200 138/20000
  983:     1 1/1 1/200 9/20000
 5090:   192 192/192 192/200 990/20000
 7502:     8 8/8 8/200 66/20000
 7563:     4 4/4 4/200 46/20000
 7637:     4 4/4 4/200 46/20000
""".strip()

PROC_KEYUSERS_INVALID_1 = """
    0    106 105/105 95/1000000 1909/25000000
  983:     1 1/1 1/200 9/20000
""".strip()

PROC_KEYUSERS_INVALID_2 = """
    0:   106: 105/105 95/1000000 1909/25000000
""".strip()

PROC_KEYUSERS_INVALID_3 = """
    0:   106 105/105 95/1000000 1909/25000000
  862:     4 4/4 4/200 40/20000
  980:     9/9 9/9 9/200 138/20000
  983:     1 1/1 1/200 9/20000
""".strip()

PROC_KEYUSERS_INVALID_4 = """
    0:   106 105/105 95/1000000 1909/25000000
  980:     9 9/9 9/2s0 138/20000
  983:     1 1/1 1/200 9/20000
""".strip()

PROC_KEYUSERS_EMPTY = """
""".strip()


def test_proc_keyusers_valid():
    key_users = ProcKeyUsers(context_wrap(PROC_KEYUSERS))
    assert len(key_users) == 8
    assert key_users[4]['uid'] == '5090'
    assert key_users[4]['usage'] == 192
    assert key_users[4]['nkeys'] == 192
    assert key_users[4]['nikeys'] == 192
    assert key_users[4]['qnkeys'] == 192
    assert key_users[4]['maxkeys'] == 200
    assert key_users[4]['qnbytes'] == 990
    assert key_users[4]['maxbytes'] == 20000
    assert key_users[6]['uid'] == '7563'


def test_proc_keyusers_invalid():
    with pytest.raises(SkipComponent):
        ProcKeyUsers(context_wrap(PROC_KEYUSERS_EMPTY))

    for input in [PROC_KEYUSERS_INVALID_1,
                    PROC_KEYUSERS_INVALID_2,
                    PROC_KEYUSERS_INVALID_3,
                    PROC_KEYUSERS_INVALID_4]:
        with pytest.raises(ParseException) as e:
            ProcKeyUsers(context_wrap(input))
        assert "Unparsable line: " in str(e)


def test_proc_keyusers_docstrings():
    env = {
        'proc_keyusers': ProcKeyUsers(context_wrap(PROC_KEYUSERS))
    }
    failed, total = doctest.testmod(proc_keyusers, globs=env)
    assert failed == 0



import pytest

from insights.core import filters
from insights.core.exceptions import SkipComponent
from insights.parsers.lsmod import LsMod
from insights.specs import Specs
from insights.specs.datasources.kernel_module_list import kernel_module_filters
from insights.tests import context_wrap

LSMOD = """
Module                  Size  Used by
igb                   249856  0
i2c_algo_bit           16384  1 igb
dca                    16384  1 igb
binfmt_misc            20480  1
tcp_diag               16384  0
udp_diag               16384  0
inet_diag              24576  2 tcp_diag,udp_diag
binfmt_misc            69632  0
udp_diag               18632  0
mfe_aac_100713183      23360   0
mfe_aac_100712495       3360   0
""".strip()

LSMOD_NO_LOADED = """
Module                  Size  Used by
tcp_diag               16384  0
""".strip()


def setup_function(func):
    if Specs.modinfo_modules in filters._CACHE:
        del filters._CACHE[Specs.modinfo_modules]
    if Specs.modinfo_modules in filters.FILTERS:
        del filters.FILTERS[Specs.modinfo_modules]

    if func is test_module_filters:
        filters.add_filter(Specs.modinfo_modules, ["udp_diag", "binfmt_misc", "wireguard"])
    if func is test_module_filters_2:
        filters.add_filter(Specs.modinfo_modules, ["mfe_aac"])
    if func is test_module_no_loaded_modules:
        filters.add_filter(Specs.modinfo_modules, ["udp_diag", "binfmt_misc", "wireguard"])
    if func is test_module_filters_empty:
        filters.add_filter(Specs.modinfo_modules, [])


def test_module_filters():
    lsmod_info = LsMod(context_wrap(LSMOD))
    broker = {LsMod: lsmod_info}
    result = kernel_module_filters(broker)
    assert 'binfmt_misc udp_diag' == result


def test_module_filters_2():
    lsmod_info = LsMod(context_wrap(LSMOD))
    broker = {LsMod: lsmod_info}
    result = kernel_module_filters(broker)
    assert 'mfe_aac_100713183 mfe_aac_100712495' == result


def test_module_filters_empty():
    lsmod_info = LsMod(context_wrap(LSMOD))
    broker = {LsMod: lsmod_info}
    with pytest.raises(SkipComponent):
        kernel_module_filters(broker)


def test_module_no_loaded_modules():
    lsmod_info = LsMod(context_wrap(LSMOD_NO_LOADED))
    broker = {LsMod: lsmod_info}
    with pytest.raises(SkipComponent):
        kernel_module_filters(broker)



import doctest
from insights.tests import context_wrap
from insights.combiners import md5check
from insights.combiners.md5check import NormalMD5
from insights.parsers.md5check import NormalMD5 as NormalMD5Parser

NORMAL_MD5_SAMPLE_1 = """
7d4855248419b8a3ce6616bbc0e58301  /etc/localtime1
""".strip()

NORMAL_MD5_SAMPLE_2 = """
d41d8cd98f00b204e9800998ecf8427e  /etc/localtime2
""".strip()


def test_normal_md5():
    parser1 = NormalMD5Parser(context_wrap(NORMAL_MD5_SAMPLE_1))
    parser2 = NormalMD5Parser(context_wrap(NORMAL_MD5_SAMPLE_2))
    md5info = NormalMD5([parser1, parser2])
    assert md5info is not None
    assert set(md5info.keys()) == set(["/etc/localtime1", "/etc/localtime2"])
    assert md5info["/etc/localtime1"] == "7d4855248419b8a3ce6616bbc0e58301"
    assert md5info["/etc/localtime2"] == "d41d8cd98f00b204e9800998ecf8427e"


def test_md5_docs():
    parser1 = NormalMD5Parser(context_wrap(NORMAL_MD5_SAMPLE_1))
    parser2 = NormalMD5Parser(context_wrap(NORMAL_MD5_SAMPLE_2))
    env = {'md5sums': NormalMD5([parser1, parser2])}
    failed, total = doctest.testmod(md5check, globs=env)
    assert failed == 0



from insights.parsers.sys_fs_cgroup_memory_tasks_number import SysFsCgroupMemoryTasksNumber
from insights.tests import context_wrap


TASKS_NUMBER = """
260
""".strip()


def test_sys_fs_cgroup_memory_tasks_number():
    sys_fs_cgroup_memory_tasks_number_content = SysFsCgroupMemoryTasksNumber(context_wrap(TASKS_NUMBER))
    assert sys_fs_cgroup_memory_tasks_number_content.number == 260



"""
Custom datasource to get all updated/created eap reports in last 24 hours
"""

import os

from insights.core.context import HostContext
from insights.core.plugins import datasource
from insights.core.exceptions import SkipComponent
from insights.specs.datasources import get_recent_files


@datasource(HostContext)
def eap_report_files(broker):
    """
    Get all updated/created eap reports in last 24 hours.

    Returns:
        function: return the list of EAP reports file in last 24 hours.
    """
    ctx = broker[HostContext]
    root = ctx.root
    EAP_REPORTS_JSON_PATH = os.path.join(root, "var/tmp/insights-runtimes/uploads/")
    final_files = get_recent_files(EAP_REPORTS_JSON_PATH, 24)
    if final_files:
        return [f[len(root):] for f in final_files]
    raise SkipComponent



"""
cpuset_cpus - File ``/sys/fs/cgroup/cpuset/cpuset.cpus``
========================================================

This parser reads the content of ``/sys/fs/cgroup/cpuset/cpuset.cpus``.
This file shows the default cgroup cpuset.cpu of system. The format
of the content is string including comma.
"""

from insights.core import CommandParser, ContainerParser
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.cpuset_cpus)
class CpusetCpus(CommandParser):
    """
    Class ``CpusetCpus`` parses the content of the ``/sys/fs/cgroup/cpuset/cpuset.cpus``.

    Attributes:
        cpu_set (list): It is used to show the list of allowed cpu.

        cpu_number (int): It is used to display the number of allowed cpu.

    A small sample of the content of this file looks like::

        0,2-4,7

    Examples:
        >>> type(cpusetinfo)
        <class 'insights.parsers.cpuset_cpus.CpusetCpus'>
        >>> cpusetinfo.cpu_set
        ['0', '2', '3', '4', '7']
        >>> cpusetinfo.cpu_number
        5
    """

    def parse_content(self, content):
        self.cpu_set = []
        self.cpu_number = 0
        values = content[0].strip().split(",")
        for value in values:
            if "-" in value:
                # Parse the value like "2-4"
                start, end = value.split("-")
                self.cpu_set.extend([str(i) for i in range(int(start), int(end) + 1)])
            else:
                self.cpu_set.append(value)
        self.cpu_number = len(self.cpu_set)


@parser(Specs.container_cpuset_cpus)
class ContainerCpusetCpus(ContainerParser, CpusetCpus):
    """
    Class ``ContainerCpusetCpus`` parses the content of the ``/sys/fs/cgroup/cpuset/cpuset.cpus`` from containers.

    Attributes:
        cpu_set (list): It is used to show the list of allowed cpu.

        cpu_number (int): It is used to display the number of allowed cpu.

    A small sample of the content of this file looks like::

        0,2-4,7

    Examples:
        >>> type(container_cpusetinfo)
        <class 'insights.parsers.cpuset_cpus.ContainerCpusetCpus'>
        >>> container_cpusetinfo.container_id
        '2869b4e2541c'
        >>> container_cpusetinfo.image
        'registry.access.redhat.com/ubi8/nginx-120'
        >>> container_cpusetinfo.cpu_set
        ['0', '2', '3', '4', '7']
        >>> container_cpusetinfo.cpu_number
        5
    """
    pass



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import rear_conf
from insights.parsers.rear_conf import RearLocalConf
from insights.tests import context_wrap

RDMA_CONFIG = """
# This file etc/rear/local.conf is intended for the user's
# manual configuration of Relax-and-Recover (ReaR).
# For configuration through packages and other automated means
# we recommend a separated file named site.conf next to this file
# and leave local.conf as is (ReaR upstream will never ship a site.conf).
# The default OUTPUT=ISO creates the ReaR rescue medium as ISO image.
# You need to specify your particular backup and restore method for your data
# as the default BACKUP=REQUESTRESTORE does not really do that (see "man rear").
# Configuration variables are documented in /usr/share/rear/conf/default.conf
# and the examples in /usr/share/rear/conf/examples/ can be used as templates.
# ReaR reads the configuration files via the bash builtin command 'source'
# so bash syntax like VARIABLE="value" (no spaces at '=') is mandatory.
# Because 'source' executes the content as bash scripts you can run commands
# within your configuration files, in particular commands to set different
# configuration values depending on certain conditions as you need like
# CONDITION_COMMAND && VARIABLE="special_value" || VARIABLE="usual_value"
# but that means CONDITION_COMMAND gets always executed when 'rear' is run
# so ensure nothing can go wrong if you run commands in configuration files.
BACKUP_RESTORE_MOVE_AWAY_FILES=( /boot/grub/grubenv /boot/grub2/grubenv )
""".strip()

REAR_CONF_EMPTY = """
""".strip()


def test_rdma_config():
    local_conf = RearLocalConf(context_wrap(RDMA_CONFIG))
    assert len(local_conf.lines) == 1
    assert local_conf.lines[0] == "BACKUP_RESTORE_MOVE_AWAY_FILES=( /boot/grub/grubenv /boot/grub2/grubenv )"


def test_rdma_config_empty():
    with pytest.raises(SkipComponent):
        RearLocalConf(context_wrap(REAR_CONF_EMPTY))


def test_rdma_config_doc():
    env = {
            'local_conf': RearLocalConf(context_wrap(RDMA_CONFIG)),
          }
    failed, total = doctest.testmod(rear_conf, globs=env)
    assert failed == 0



"""
LsOcpCniOpenshiftSdn - command ``ls -l /var/lib/cni/networks/openshift-sdn``
============================================================================

The ``ls -l /var/lib/cni/networks/openshift-sdn`` command is used to return the count of cni files
and also could provide information for the listing of the ``/var/lib/cni/networks/openshift-sdn``
directory. See ``FileListing`` class for additional information.
Sample input is shown in the Examples. See ``FileListing`` class for
additional information.

Sample directory list::

    total 52
    -rw-r--r--. 1 root root 64 Aug  5 23:26 10.130.0.102
    -rw-r--r--. 1 root root 64 Aug  5 23:26 10.130.0.103
    -rw-r--r--. 1 root root 64 Aug  6 22:52 10.130.0.116
    -rw-r--r--. 1 root root 64 Aug  6 22:52 10.130.0.117
    -rw-r--r--. 1 root root 64 Aug  5 06:59 10.130.0.15
    -rw-r--r--. 1 root root 64 Aug  5 07:02 10.130.0.20
    -rw-r--r--. 1 root root 12 Aug  6 22:52 last_reserved_ip.0

Examples:

    >>> ls_ocp_cni_openshift_sdn.files_of("/var/lib/cni/networks/openshift-sdn")
    ['10.130.0.102', '10.130.0.103', '10.130.0.116', '10.130.0.117', '10.130.0.15', '10.130.0.20', 'last_reserved_ip.0']
"""

from .. import FileListing
from .. import parser, CommandParser
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.ls_ocp_cni_openshift_sdn)
class LsOcpCniOpenshiftSdn(CommandParser, FileListing):
    """
    .. warning::
        This class is deprecated and will be removed from 3.5.0.
        Please use the :class:`insights.parsers.ls.LSla` instead.

    Parses output of ``ls -l /var/lib/cni/networks/openshift-sdn`` command.
    """
    def __init__(self, *args, **kwargs):
        deprecated(LsOcpCniOpenshiftSdn, "Please use the :class:`insights.parsers.ls.LSla` instead.", "3.5.0")
        super(LsOcpCniOpenshiftSdn, self).__init__(*args, **kwargs)



"""
QemuXML - file ``/etc/libvirt/qemu/*.xml`` and ``/var/run/libvirt/qemu/*.xml``
==============================================================================

Parsers provided by this module are:

QemuXML - file ``/etc/libvirt/qemu/*.xml``
------------------------------------------

VarQemuXML - file ``/var/run/libvirt/qemu/*.xml``
-------------------------------------------------

OpenStackInstanceXML - file ``/etc/libvirt/qemu/*.xml``
-------------------------------------------------------
"""
from insights.components.openstack import IsOpenStackCompute
from insights.specs import Specs
from .. import XMLParser, parser


class BaseQemuXML(XMLParser):
    """Base class for parsing Qemu XML files. It uses ``XMLParser`` mixin
    class.

    Attributes:
        vm_name(str): Name of VM

    """
    def parse_content(self, content):
        super(BaseQemuXML, self).parse_content(content)

        # Setting vm_name attribute
        if self.dom is not None and len(self.get_elements('./domain/name')):
            self.vm_name = self.get_elements('./domain/name')[0].text
        else:
            self.vm_name = self.data.get('name', None)

    def parse_dom(self):
        """Parse xml information in :attr:`data` and return.

        Returns:
            dict: Parsed xml data. An empty dictionary when content is blank.
        """
        if self.dom is None:
            return {}

        domain = {}
        for child in self.dom:
            if len(child) == 0:
                domain[child.tag] = child.text
            else:
                domain[child.tag] = [c.items() for c in child]
        return domain


@parser(Specs.qemu_xml)
class QemuXML(BaseQemuXML):
    """This class parses xml files under ``/etc/libvirt/qemu/`` using
    ``BaseQemuXML`` base parser.

    Sample file::

        <!--
        WARNING: THIS IS AN AUTO-GENERATED FILE. CHANGES TO IT ARE LIKELY TO BE
        OVERWRITTEN AND LOST. Changes to this xml configuration should be made using:
          virsh edit 05-s00c06h0
        or other application using the libvirt API.
        -->

        <domain type='kvm'>
          <name>05-s00c06h0</name>
          <uuid>02cf0bba-2bd6-11e7-8337-e4115b9a50d0</uuid>
          <memory unit='KiB'>12582912</memory>
          <currentMemory unit='KiB'>12582912</currentMemory>
          <vcpu placement='static'>4</vcpu>
          <cputune>
            <vcpupin vcpu='0' cpuset='1'/>
            <vcpupin vcpu='1' cpuset='2'/>
            <vcpupin vcpu='2' cpuset='3'/>
            <vcpupin vcpu='3' cpuset='4'/>
            <emulatorpin cpuset='1-4'/>
          </cputune>
          <numatune>
            <memory mode='strict' nodeset='0-1'/>
            <memnode cellid='0' mode='strict' nodeset='0'/>
            <memnode cellid='1' mode='strict' nodeset='1'/>
          </numatune>
          <os>
            <type arch='x86_64' machine='pc-i440fx-rhel7.0.0'>hvm</type>
            <boot dev='hd'/>
            <boot dev='network'/>
            <bootmenu enable='yes' timeout='1000'/>
            <bios useserial='yes' rebootTimeout='0'/>
          </os>
          <features>
            <acpi/>
            <apic/>
            <pae/>
          </features>
          <cpu>
            <numa>
              <cell id='0' cpus='0-1' memory='6291456' unit='KiB'/>
              <cell id='1' cpus='2-3' memory='6291456' unit='KiB'/>
            </numa>
          </cpu>
          <clock offset='utc'/>
          <on_poweroff>destroy</on_poweroff>
          <on_reboot>restart</on_reboot>
          <on_crash>restart</on_crash>
          <devices>
            <emulator>/usr/libexec/qemu-kvm</emulator>
            <disk type='file' device='disk'>
              <driver name='qemu' type='raw' cache='none' io='threads'/>
              <source file='/var/lib/libvirt/images/05-s00c06h0_1.img'/>
              <target dev='vda' bus='virtio'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x05' function='0x0'/>
            </disk>
            <controller type='usb' index='0'>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x01' function='0x2'/>
            </controller>
            <controller type='pci' index='0' model='pci-root'/>
            <controller type='virtio-serial' index='0'>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x03' function='0x0'/>
            </controller>
            <interface type='hostdev' managed='yes'>
              <mac address='b2:59:73:15:00:00'/>
              <source>
                <address type='pci' domain='0x0000' bus='0x04' slot='0x10' function='0x0'/>
              </source>
              <rom bar='on' file='/opt/vcp/share/ipxe/808610ed.rom'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x04' function='0x0'/>
            </interface>
            <interface type='hostdev' managed='yes'>
              <mac address='b2:59:73:15:00:01'/>
              <source>
                <address type='pci' domain='0x0000' bus='0x04' slot='0x10' function='0x1'/>
              </source>
              <rom bar='on' file='/opt/vcp/share/ipxe/808610ed.rom'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x08' function='0x0'/>
            </interface>
            <serial type='pty'>
              <target port='0'/>
            </serial>
            <console type='pty'>
              <target type='serial' port='0'/>
            </console>
            <channel type='pipe'>
              <source path='/var/lib/libvirt/qemu/channels/FROM-05-s00c06h0'/>
              <target type='virtio' name='virtio2host'/>
              <address type='virtio-serial' controller='0' bus='0' port='1'/>
            </channel>
            <channel type='pipe'>
              <source path='/var/lib/libvirt/qemu/channels/HGC-05-s00c06h0'/>
              <target type='virtio' name='virtio_host_guest_check'/>
              <address type='virtio-serial' controller='0' bus='0' port='2'/>
            </channel>
            <input type='mouse' bus='ps2'/>
            <input type='keyboard' bus='ps2'/>
            <graphics type='vnc' port='-1' autoport='yes'>
              <listen type='address'/>
            </graphics>
            <video>
              <model type='cirrus' vram='16384' heads='1' primary='yes'/>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x02' function='0x0'/>
            </video>
            <watchdog model='i6300esb' action='reset'>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x07' function='0x0'/>
            </watchdog>
            <memballoon model='virtio'>
              <address type='pci' domain='0x0000' bus='0x00' slot='0x06' function='0x0'/>
            </memballoon>
          </devices>
        </domain>

    Examples:
        >>> xml_numa.file_name == 'vm.xml'
        True
        >>> xml_numa.vm_name == '05-s00c06h0'
        True
        >>> memnode = xml_numa.get_elements('./numatune/memnode', None)
        >>> len(memnode[0].items()) == 3
        True
        >>> len(memnode[1].items()) == 3
        True
        >>> memnode[0].get('cellid') == '0'
        True
        >>> memnode[1].get('mode') == 'strict'
        True
    """
    pass


@parser(Specs.var_qemu_xml)
class VarQemuXML(BaseQemuXML):
    """This class parses xml files under ``/var/run/libvirt/qemu/`` using
    ``BaseQemuXML`` base parser.

    Sample file::

        <!--
        WARNING: THIS IS AN AUTO-GENERATED FILE. CHANGES TO IT ARE LIKELY TO BE
        OVERWRITTEN AND LOST. Changes to this xml configuration should be made using:
          virsh edit test-idm-client-ccveu-net
        or other application using the libvirt API.
        -->

        <domstatus state='running' reason='unpaused' pid='17150'>
          <monitor path='/var/lib/libvirt/qemu/domain-59-test-idm-client-ccve/monitor.sock' json='1' type='unix'/>
          <vcpus>
            <vcpu id='0' pid='17156'/>
          </vcpus>
          <qemuCaps>
            <flag name='kvm'/>
            <flag name='mem-path'/>
          </qemuCaps>
          <devices>
            <device alias='balloon0'/>
          </devices>
          <libDir path='/var/lib/libvirt/qemu/domain-59-test-idm-client-ccve'/>
          <domain type='kvm' id='59'>
            <name>test-idm-client-ccveu-net</name>
            <uuid>78177d07-ac0e-4057-b1de-9ccd66cbc3d7</uuid>
            <metadata xmlns:ovirt="http://ovirt.org/vm/tune/1.0">
              <ovirt:qos/>
            </metadata>
            <maxMemory slots='16' unit='KiB'>4294967296</maxMemory>
            <memory unit='KiB'>2097152</memory>
            <os>
              <type arch='x86_64' machine='pc-i440fx-rhel7.2.0'>hvm</type>
              <bootmenu enable='yes' timeout='10000'/>
              <smbios mode='sysinfo'/>
            </os>
            <devices>
              <emulator>/usr/libexec/qemu-kvm</emulator>
              <disk type='file' device='cdrom'>
                <driver name='qemu' type='raw'/>
                <source startupPolicy='optional'/>
              </disk>
            </devices>
          </domain>
        </domstatus>
    """
    pass


@parser(Specs.qemu_xml, IsOpenStackCompute)
class OpenStackInstanceXML(BaseQemuXML):
    """Parse OpenStack instances metadata based on the class ``BaseQemuXML``.

    This parser depends on ``insights.components.openstack.IsOpenStackCompute``
    and will be fired only if the dependency is met.

    Sample metadata section in the XML file::

        <metadata>
        <nova:instance xmlns:nova="http://openstack.org/xmlns/libvirt/nova/1.0">
          <nova:package version="14.0.3-8.el7ost"/>
          <nova:name>django_vm_001</nova:name>
          <nova:creationTime>2017-10-09 08:51:28</nova:creationTime>
          <nova:flavor name="vpc1-cf1-foo-bar">
            <nova:memory>8096</nova:memory>
            <nova:disk>10</nova:disk>
            <nova:swap>0</nova:swap>
            <nova:ephemeral>0</nova:ephemeral>
            <nova:vcpus>4</nova:vcpus>
          </nova:flavor>
          <nova:owner>
            <nova:user uuid="96e9d2b749ea48fcb5a911e6f0e144f2">django_user_01</nova:user>
            <nova:project uuid="5a50e9d0d19746158958be0c759793fb">vpcdi1</nova:project>
          </nova:owner>
          <nova:root type="image" uuid="1a05a423-dfae-428a-ae54-1614d8024e76"/>
        </nova:instance>
        </metadata>

    Examples:
        >>> rhosp_xml.domain_name
        'instance-000008d6'
        >>> rhosp_xml.nova.get('version')
        '14.0.3-8.el7ost'
        >>> rhosp_xml.nova.get('instance_name')
        'django_vm_001'
        >>> rhosp_xml.nova.get('user')
        'django_user_01'
        >>> rhosp_xml.nova.get('root_disk_type')
        'image'
        >>> rhosp_xml.nova.get('flavor_vcpus')
        '4'

    Attributes:
        domain_name(str): XML domain name.
        nova(dict): OpenStack Compute Metadata.
    """
    def parse_content(self, content):
        super(OpenStackInstanceXML, self).parse_content(content)
        self.domain_name = self.file_name.strip('.xml')

        # Parse OpenStack Compute(Nova) metadata
        ns = {'nova': 'http://openstack.org/xmlns/libvirt/nova/1.0'}
        self.nova = {}
        if self.dom is not None and len(self.get_elements('./metadata')):
            for i in self.get_elements('./metadata')[0]:
                self.nova['version'] = i.findall('nova:package', ns)[0].get('version')

                # Instance meta
                self.nova['instance_name'] = i.findall('nova:name', ns)[0].text
                self.nova['created'] = i.findall('nova:creationTime', ns)[0].text

                # Flavor meta
                if i.findall('nova:flavor', ns):
                    flavor = i.findall('nova:flavor', ns)[0]
                    self.nova['flavor_name'] = flavor.get('name')
                    self.nova['flavor_memory'] = flavor.findall('nova:memory', ns)[0].text
                    self.nova['flavor_disk'] = flavor.findall('nova:disk', ns)[0].text
                    self.nova['flavor_swap'] = flavor.findall('nova:swap', ns)[0].text
                    self.nova['flavor_ephemeral'] = flavor.findall('nova:ephemeral', ns)[0].text
                    self.nova['flavor_vcpus'] = flavor.findall('nova:vcpus', ns)[0].text

                # Owner meta
                if i.findall('nova:owner', ns):
                    owner = i.findall('nova:owner', ns)[0]
                    user = owner.findall('nova:user', ns)[0]
                    project = owner.findall('nova:project', ns)[0]
                    self.nova['user'] = user.text
                    self.nova['user_uuid'] = user.get('uuid')
                    self.nova['project'] = project.text
                    self.nova['project_uuid'] = project.get('uuid')

                # root disk meta
                if i.findall('nova:root', ns):
                    root = i.findall('nova:root', ns)[0]
                    self.nova['root_disk_type'] = root.get('type')
                    self.nova['root_disk_uuid'] = root.get('uuid')



"""
Libvirtd Logs
=============

This module contains the following parsers:

LibVirtdLog - file ``/var/log/libvirt/libvirtd.log``
----------------------------------------------------
LibVirtdQemuLog - file ``/var/log/libvirt/qemu/*.log``
------------------------------------------------------
"""
from insights.specs import Specs
from insights import LogFileOutput, parser


@parser(Specs.libvirtd_log)
class LibVirtdLog(LogFileOutput):
    """
    Parse the ``/var/log/libvirt/libvirtd.log`` log file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`

    Sample input::

        2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1174 : trying driver 0 (Test) ...
        2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1180 : driver 0 Test returned DECLINED
        2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1174 : trying driver 1 (ESX) ...
        2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1180 : driver 1 ESX returned DECLINED
        2013-10-23 17:32:19.909+0000: 14069: debug : do_open:1174 : trying driver 2 (remote) ...
        2013-10-23 17:32:19.957+0000: 14069: error : virNetTLSContextCheckCertDN:418 : Certificate [session] owner does not match the hostname AA.BB.CC.DD <============= IP Address
        2013-10-23 17:32:19.957+0000: 14069: warning : virNetTLSContextCheckCertificate:1102 : Certificate check failed Certificate [session] owner does not match the hostname AA.BB.CC.DD
        2013-10-23 17:32:19.957+0000: 14069: error : virNetTLSContextCheckCertificate:1105 : authentication failed: Failed to verify peer's certificate

    Examples:

        >>> "Certificate check failed Certificate" in libvirtd_log
        True
        >>> len(libvirtd_log.lines) # All lines, before filtering
        8
        >>> len(libvirtd_log.get('NetTLSContext')) # After filtering
        3
    """
    pass


@parser(Specs.libvirtd_qemu_log)
class LibVirtdQemuLog(LogFileOutput):
    """
        Parse the ``/var/log/libvirt/qemu/*.log`` log file.

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`

    Sample input from file /var/log/libvirt/qemu/bb912729-fa51-443b-bac6-bf4c795f081d.log::

        2019-06-04 05:33:22.280743Z qemu-kvm: -vnc 10.xxx.xxx.xxx:0: Failed to start VNC server: Failed to bind socket: Cannot assign requested address
        2019-06-04 05:33:2.285+0000: shutting down

    Examples:

        >>> from datetime import datetime
        >>> "shutting down" in libvirtd_qemu_log
        True
        >>> len(list(libvirtd_qemu_log.get_after(datetime(2019, 4, 26, 6, 55, 20))))
        2
        >>> libvirtd_qemu_log.file_name.strip('.log')  # Instance UUID
        'bb912729-fa51-443b-bac6-bf4c795f081d'
    """
    pass



from insights.specs import Specs
from insights.tests import InputData, archive_provider, RHEL7, MAKE_NONE_RESULT
from insights.plugins import returns_none


@archive_provider(returns_none.report_make_none)
def integration_tests_1():
    input_data = InputData("test_return_make_none_1")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, None

    input_data = InputData("test_return_make_none_2")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, MAKE_NONE_RESULT


@archive_provider(returns_none.report_none)
def integration_tests_2():
    input_data = InputData("test_return_none_1")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, None

    input_data = InputData("test_return_none_2")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, MAKE_NONE_RESULT


@archive_provider(returns_none.report_skip_exception)
def integration_tests_3():
    input_data = InputData("test_return_skip_exception_1")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, None

    input_data = InputData("test_return_skip_exception_2")
    input_data.add(Specs.redhat_release, RHEL7)
    yield input_data, None


def test_integration_1(run_rule):
    input_data = InputData("test_return_make_none_1_1")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_make_none, input_data)
    assert result is None

    input_data = InputData("test_return_make_none_2_2")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_make_none, input_data, return_make_none=True)
    assert result == MAKE_NONE_RESULT


def test_integration_2(run_rule):
    input_data = InputData("test_return_none_2_1")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_none, input_data)
    assert result is None

    input_data = InputData("test_return_none_2_2")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_none, input_data, return_make_none=True)
    assert result == MAKE_NONE_RESULT


def test_integration_3(run_rule):
    input_data = InputData("test_return_skip_exception_2_1")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_skip_exception, input_data)
    assert result is None

    input_data = InputData("test_return_skip_exception_2_2")
    input_data.add(Specs.redhat_release, RHEL7)
    result = run_rule(returns_none.report_skip_exception, input_data, return_make_none=True)
    assert result is None



import pytest

from insights.tests import context_wrap
from insights.parsers.postgresql_conf import PostgreSQLConf

postgresql_conf_cnt = """

# -----------------------------
# PostgreSQL configuration file
# -----------------------------
#
# This file consists of lines of the form:
#
#   name = value
#
# (The "=" is optional.)  Whitespace may be used.  Comments are introduced with
# "#" anywhere on a line.  The complete list of parameter names and allowed
# values can be found in the PostgreSQL documentation.
#
# The commented-out settings shown in this file represent the default values.
# Re-commenting a setting is NOT sufficient to revert it to the default value;
# you need to reload the server.
#
# This file is read on server startup and when the server receives a SIGHUP
# signal.  If you edit the file on a running system, you have to SIGHUP the
# server for the changes to take effect, or use "pg_ctl reload".  Some
# parameters, which are marked below, require a server shutdown and restart to
# take effect.
#
# Any parameter can also be given as a command-line option to the server, e.g.,
# "postgres -c log_connections=on".  Some parameters can be changed at run time
# with the "SET" SQL command.
#
# Memory units:  kB = kilobytes        Time units:  ms  = milliseconds
#                MB = megabytes                     s   = seconds
#                GB = gigabytes                     min = minutes
#                                                   h   = hours
#                                                   d   = days


#------------------------------------------------------------------------------
# FILE LOCATIONS
#------------------------------------------------------------------------------

# The default values of these variables are driven from the -D command-line
# option or PGDATA environment variable, represented here as ConfigDir.

#data_directory = 'ConfigDir'		# use data in another directory
#hba_file = 'ConfigDir/pg_hba.conf'	# host-based authentication file
#ident_file = 'ConfigDir/pg_ident.conf'	# ident configuration file

# If external_pid_file is not explicitly set, no extra PID file is written.
#external_pid_file = '(none)'		# write an extra PID file


#------------------------------------------------------------------------------
# CONNECTIONS AND AUTHENTICATION
#------------------------------------------------------------------------------

# - Connection Settings -

#listen_addresses = 'localhost'		# what IP address(es) to listen on;
#port = 5432				# (change requires restart)
### next line has been commented out by spacewalk-setup-postgresql ###
##max_connections = 100			# (change requires restart)
# Note:  Increasing max_connections costs ~400 bytes of shared memory per
# connection slot, plus lock space (see max_locks_per_transaction).
#superuser_reserved_connections = 3	# (change requires restart)
#unix_socket_directory = ''		# (change requires restart)
#unix_socket_group = ''			# (change requires restart)
#unix_socket_permissions = 0777		# begin with 0 to use octal notation
#bonjour_name = ''			# defaults to the computer name

# - Security and Authentication -

#authentication_timeout = 1min		# 1s-600s
#ssl = off				# (change requires restart)
#ssl_ciphers = 'ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH'	# allowed SSL ciphers
#ssl_renegotiation_limit = 512MB	# amount of data between renegotiations
#password_encryption = on
#db_user_namespace = off

# Kerberos and GSSAPI
#krb_server_keyfile = ''
#krb_srvname = 'postgres'		# (Kerberos only)
#krb_caseins_users = off

# - TCP Keepalives -
# see "man 7 tcp" for details

#tcp_keepalives_idle = 0		# TCP_KEEPIDLE, in seconds;
#tcp_keepalives_interval = 0		# TCP_KEEPINTVL, in seconds;
#tcp_keepalives_count = 0		# TCP_KEEPCNT;

#------------------------------------------------------------------------------
# RESOURCE USAGE (except WAL)
#------------------------------------------------------------------------------

# - Memory -

### next line has been commented out by spacewalk-setup-postgresql ###
##shared_buffers = 32MB			# min 128kB
#temp_buffers = 8MB			# min 800kB
#max_prepared_transactions = 0		# zero disables the feature
# Note:  Increasing max_prepared_transactions costs ~600 bytes of shared memory
# per transaction slot, plus lock space (see max_locks_per_transaction).
# It is not advisable to set max_prepared_transactions nonzero unless you
# actively intend to use prepared transactions.
#work_mem = 1MB				# min 64kB
#maintenance_work_mem = 16MB		# min 1MB
#max_stack_depth = 2MB			# min 100kB

# - Kernel Resource Usage -

#max_files_per_process = 1000		# min 25
#shared_preload_libraries = ''		# (change requires restart)

# - Cost-Based Vacuum Delay -

#vacuum_cost_delay = 0ms		# 0-100 milliseconds
#vacuum_cost_page_hit = 1		# 0-10000 credits
#vacuum_cost_page_miss = 10		# 0-10000 credits
#vacuum_cost_page_dirty = 20		# 0-10000 credits
#vacuum_cost_limit = 200		# 1-10000 credits

# - Background Writer -

#bgwriter_delay = 200ms			# 10-10000ms between rounds
#bgwriter_lru_maxpages = 100		# 0-1000 max buffers written/round
#bgwriter_lru_multiplier = 2.0		# 0-10.0 multipler on buffers scanned/round

# - Asynchronous Behavior -

#effective_io_concurrency = 1		# 1-1000. 0 disables prefetching

# These are only used if logging_collector is on:
log_directory = 'pg_log'		# directory where log files are written,
log_filename = 'postgresql-%a.log'	# log file name pattern,
log_truncate_on_rotation = on		# If on, an existing log file of the
checkpoint_completion_target = 0.7
checkpoint_segments = 8
effective_cache_size = 1152MB
log_line_prefix = '%m '
maintenance_work_mem = 96MB
max_connections = 600
shared_buffers = 384MB
wal_buffers = 4MB
work_mem = 2560kB

password_encryption on
db_user_namespace = off

bgwriter_delay = 200ms			# 10-10000ms between rounds
checkpoint_timeout = 5min
tcp_keepalives_interval 300

max_stack_depth = 2048576       # Test of as_memory_bytes with string of digits

test_strange_quoting '''strange quoting\\''
""".strip()


def test_postgresql_conf():
    result = PostgreSQLConf(context_wrap(postgresql_conf_cnt))
    assert result.get("checkpoint_segments") == "8"
    # The bit before the hash mark should still be treated as valid:
    assert result.get("log_filename") == "postgresql-%a.log"
    # Quoting allows spaces at beginning or end of value
    assert result.get("log_line_prefix") == "%m "
    # Equals signs are optional
    assert result.get("password_encryption") == "on"
    # Values can include a quote with '' or \\' - both work.
    assert result.get("test_strange_quoting") == "'strange quoting'"

    # Default value tests
    # get method from LegacyItemAccess
    assert result.get(None) is None
    assert result.get('') is None
    assert 'listen_addresses' not in result
    assert result.get('listen_addresses', 'localhost') == 'localhost'


def test_postgresql_conf_conversions():
    result = PostgreSQLConf(context_wrap(postgresql_conf_cnt))
    assert result.as_duration('bgwriter_delay') == 0.2
    assert result.as_duration('checkpoint_timeout') == 300
    assert result.as_duration('tcp_keepalives_interval') == 300
    # Default value tests do conversions as well
    assert result.as_duration(None) is None
    assert 'vacuum_cost_delay' not in result
    assert result.as_duration('vacuum_cost_delay', '200ms') == 0.2
    assert result.as_duration('tcp_keepalives_idle', '0') == 0
    assert result.as_duration('tcp_keepalives_idle', 0) == 0

    assert result.as_boolean('password_encryption')
    assert not result.as_boolean('db_user_namespace')
    # Default value tests do conversions as well
    assert result.as_boolean(None) is None
    assert result.as_boolean('no_such_property', True)
    assert 'krb_caseins_users' not in result
    assert not result.as_boolean('krb_caseins_users', 'no')

    assert result.as_memory_bytes('work_mem') == 2560 * 1024
    assert result.as_memory_bytes('wal_buffers') == 4 * 1048576
    # No scaling necessary if no suffix but conversion to int done
    assert result.as_memory_bytes('max_stack_depth') == 2048576
    # Default value tests do conversions as well
    assert result.as_memory_bytes(None) is None
    assert 'temp_buffers' not in result
    assert result.as_memory_bytes('temp_buffers', '8MB') == 8192 * 1024
    assert result.as_memory_bytes('temp_buffers', '8388608') == 8192 * 1024
    assert result.as_memory_bytes('temp_buffers', 8388608) == 8192 * 1024


def test_postgresql_conf_conversion_errors():
    result = PostgreSQLConf(context_wrap(postgresql_conf_cnt))
    # Test that we raise the right errors for bad conversions
    # Can't chain them because the raised error aborts further checks.
    with pytest.raises(ValueError):
        assert result.as_duration('log_filename')
    with pytest.raises(ValueError):
        assert result.as_duration('db_user_namespace')
    with pytest.raises(ValueError):
        assert result.as_boolean('log_directory')
    with pytest.raises(ValueError):
        assert result.as_boolean('checkpoint_segments')
    with pytest.raises(ValueError):
        assert result.as_memory_bytes('log_line_prefix')
    with pytest.raises(ValueError):
        assert result.as_memory_bytes('checkpoint_timeout')



"""
Custom datasources related to ``md5``
"""
from insights.core.context import HostContext
from insights.core.plugins import datasource


#
# This function is quite simple, do not need to test
#
@datasource(HostContext)
def files(broker):
    """
    Return a list of files to be processed by the ``md5chk_files`` spec
    """
    return [
        "/etc/pki/product/69.pem",
        "/etc/pki/product-default/69.pem",
        "/usr/lib/libsoftokn3.so",
        "/usr/lib64/libsoftokn3.so",
        "/usr/lib/libfreeblpriv3.so",
        "/usr/lib64/libfreeblpriv3.so"
    ]



from insights.tests import context_wrap
from insights.parsers.sysconfig import PcsdSysconfig

PCSD_SYSCONFIG = """
# Set PCSD_DEBUG to true for advanced pcsd debugging information
PCSD_DEBUG=false
# Set DISABLE_GUI to true to disable GUI frontend in pcsd
PCSD_DISABLE_GUI=false
# Set web UI sesions lifetime in seconds
PCSD_SESSION_LIFETIME=3600
# List of IP addresses pcsd should bind to delimited by ',' character
#PCSD_BIND_ADDR='::'
""".strip()


def test_sysconfig_virt_who():
    result = PcsdSysconfig(context_wrap(PCSD_SYSCONFIG))
    assert result["PCSD_DEBUG"] == 'false'
    assert result["PCSD_SESSION_LIFETIME"] == '3600'
    assert "NOSUCHOPTIONS" not in result
